import logging
import math
import os
import shutil
from multiprocessing import Manager
from typing import Any, Callable, Union

from joblib import Parallel, delayed
from tqdm import tqdm

from torcheeg.io import EEGSignalIO, MetaInfoIO

log = logging.getLogger(__name__)


def set_records(self, **kwargs):
    root_path = kwargs.pop('root_path', '.')  # str
    num_samples_per_worker = kwargs.pop('num_samples_per_worker', '.')  # str

    existing_meta_info_io_path = os.path.join(root_path, 'info.csv')
    existing_info_io = MetaInfoIO(existing_meta_info_io_path)
    df = existing_info_io.read_all()

    block_list = list()
    num_block_list = math.ceil(len(df) / num_samples_per_worker)
    for i in range(num_block_list):
        block_list.append(
            [i * num_samples_per_worker, (i + 1) * num_samples_per_worker])
    return block_list


def process_record(block,
                   io_path: str = None,
                   io_size: int = 10485760,
                   io_mode: str = 'lmdb',
                   lock: Any = None,
                   **kwargs):
    start_id, end_id = block

    transform = kwargs.pop('transform', None)
    root_path = kwargs.pop('root_path', '.')

    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    existing_meta_info_io_path = os.path.join(io_path, 'info.csv')
    existing_eeg_signal_io_path = os.path.join(root_path, 'eeg')

    existing_info_io = MetaInfoIO(existing_meta_info_io_path)
    existing_eeg_io = EEGSignalIO(existing_eeg_signal_io_path,
                                  io_size=io_size,
                                  io_mode=io_mode)

    df = existing_info_io.read_all()
    chunk = df[start_id:end_id]

    eeg_io = EEGSignalIO(eeg_signal_io_path, io_size=io_size, io_mode=io_mode)

    last_baseline_id = None
    last_baseline_sample = None

    for i, row in chunk.iterrows():
        clip_id = row['clip_id']
        clip_sample = existing_eeg_io.read_eeg(clip_id)

        if 'baseline_id' in row:
            baseline_id = row['baseline_id']

            if last_baseline_id == baseline_id:
                trial_baseline_sample = last_baseline_sample
            else:
                trial_baseline_sample = existing_eeg_io.read_eeg(baseline_id)

            if transform is not None:
                t = transform(eeg=clip_sample, baseline=trial_baseline_sample)
            else:
                t = dict(eeg=clip_sample, baseline=trial_baseline_sample)

            t_eeg = t['eeg']
            t_baseline = t['baseline']

            if not last_baseline_id == baseline_id:
                with lock:
                    eeg_io.write_eeg(t_baseline, baseline_id)

                last_baseline_id = baseline_id
                last_baseline_sample = trial_baseline_sample

            with lock:
                eeg_io.write_eeg(t_eeg, clip_id)

        else:
            if transform is not None:
                t = transform(eeg=clip_sample)
            else:
                t = dict(eeg=clip_sample)
            t_eeg = t['eeg']

            with lock:
                eeg_io.write_eeg(t_eeg, clip_id)


class MockLock():
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        pass


def from_existing(dataset: Any,
                  io_path: str,
                  transform: Union[None, Callable] = None,
                  num_worker: int = 0,
                  verbose: bool = True,
                  io_size: int = 10485760,
                  io_mode: str = 'lmdb',
                  num_samples_per_worker: int = 100,
                  in_memory: bool = False,
                  **kwargs):
    '''
    To create new datasets from existing databases, which is often used where new transformations are expected to be applied to already processed datasets. 

    Args:
        dataset (Basedataset): The existing database, which can be obtained through :obj:`DEAPDataset`, :obj:`SEEDDataset`, and so on.
        io_path (str): The path to generated unified data IO, cached as an intermediate result. (default: :obj:`./io/deap`)
        transform (Callable, optional): The transformation of the EEG signals and baseline EEG signals. The input is a :obj:`np.ndarray`, and the ouput is used as the first and second value of each element in the dataset. It is executed before generating IO intermediate results. (default: :obj:`None`)
        num_worker (int): Number of subprocesses to use for data loading. 0 means that the data will be loaded in the main process. (default: :obj:`0`)
        verbose (bool): Whether to display logs during processing, such as progress bars, etc. (default: :obj:`True`)
        verbose (bool): Whether to display logs during processing, such as progress bars, etc. (default: :obj:`True`)
        io_size (int): Maximum size database may grow to; used to size the memory mapping. If database grows larger than ``map_size``, an exception will be raised and the user must close and reopen. (default: :obj:`10485760`)
        io_mode (str): Storage mode of EEG signal. When io_mode is set to :obj:`lmdb`, TorchEEG provides an efficient database (LMDB) for storing EEG signals. LMDB may not perform well on limited operating systems, where a file system based EEG signal storage is also provided. When io_mode is set to :obj:`pickle`, pickle-based persistence files are used. (default: :obj:`lmdb`)
        num_samples_per_worker (int): The number of samples processed by each process. Once the specified number of samples are processed, the process will be destroyed and new processes will be created to perform new tasks. (default: :obj:`100`)
        in_memory (bool): Whether to load the entire dataset into memory. If :obj:`in_memory` is set to True, then the first time an EEG sample is read, the entire dataset is loaded into memory for subsequent retrieval. Otherwise, the dataset is stored on disk to avoid the out-of-memory problem. (default: :obj:`False`)    
    '''
    root_path = dataset.io_path
    params = {
        'root_path': root_path,
        'io_path': io_path,
        'transform': transform,
        'num_worker': num_worker,
        'verbose': verbose,
        'io_size': io_size,
        'io_mode': io_mode,
        'num_samples_per_worker': num_samples_per_worker,
        'in_memory': in_memory,
    }
    params.update(kwargs)

    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = eeg_signal_io_path = os.path.join(io_path, 'eeg')

    exist_io = os.path.exists(meta_info_io_path) and os.path.exists(
        eeg_signal_io_path)

    if not exist_io:
        print(
            f'dataset does not exist at path {io_path}, generating files to path...'
        )
        os.makedirs(io_path, exist_ok=True)

        # # init sub-folders
        shutil.copy(os.path.join(root_path, 'info.csv'),
                    os.path.join(io_path, 'info.csv'))
        EEGSignalIO(eeg_signal_io_path, io_size=io_size, io_mode=io_mode)

        if num_worker == 0:
            lock = MockLock()  # do nothing, just for compatibility
            for block in tqdm(set_records(**params),
                              disable=not verbose,
                              desc="[PROCESS]"):
                process_record(block=block, lock=lock, **params)
        else:
            # lock for lmdb writter, LMDB only allows single-process writes
            manager = Manager()
            lock = manager.Lock()

            Parallel(n_jobs=num_worker)(
                delayed(process_record)(block=block, lock=lock, **params)
                for block in tqdm(set_records(**params),
                                  disable=not verbose,
                                  desc="[PROCESS]"))
    else:
        print(
            f'dataset already exists at path {io_path}, reading from path...')

    return type(dataset)(io_path=io_path,
                         offline_transform=transform,
                         num_worker=num_worker,
                         verbose=verbose,
                         io_size=io_size,
                         in_memory=in_memory,
                         **kwargs)



from .band import *
from .concatenate import *
from .normalize import *
from .pick import *
from .to import *
from .spectrum import *
from .coefficient import *
from .correlation import *
from .band_pyeeg import *
from .downsample import *
from .rearrange import *
from .flatten import *


import os
import random
import shutil
import unittest

import numpy as np
import torch

from torcheeg.io import EEGSignalIO


class TestEEGSignalIO(unittest.TestCase):
    def setUp(self):
        shutil.rmtree('./tmp_out/')
        os.mkdir('./tmp_out/')

    def test_init(self):
        io_io_path = f'./tmp_out/{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'

        io_mode_space = ['pickle', 'lmdb']
        for io_mode in io_mode_space:
            io = EEGSignalIO(io_path=io_io_path, io_mode=io_mode)
            self.assertEqual(len(io), 0)

            self.assertTrue(os.path.exists(io_io_path))
            io = EEGSignalIO(io_path=io_io_path, io_mode=io_mode)
            self.assertEqual(len(io), 0)

    def test_write_eeg(self):
        io_io_path = f'./tmp_out/{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'

        io_mode_space = ['pickle', 'lmdb']
        for io_mode in io_mode_space:
            io = EEGSignalIO(io_path=io_io_path, io_mode=io_mode)

            eeg = np.random.randn(32, 128)
            io.write_eeg(eeg)

            eeg = np.random.randn(32, 128)
            io.write_eeg(eeg)

    def test_len(self):
        io_io_path = f'./tmp_out/{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'

        io_mode_space = ['pickle', 'lmdb']
        for io_mode in io_mode_space:
            io = EEGSignalIO(io_path=io_io_path, io_mode=io_mode)

            eeg = np.random.randn(32, 128)
            io.write_eeg(eeg)

            eeg = np.random.randn(32, 128)
            io.write_eeg(eeg)

            self.assertEqual(len(io), 2)

    def test_read_eeg(self):
        io_io_path = f'./tmp_out/{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'

        io_mode_space = ['pickle', 'lmdb']
        for io_mode in io_mode_space:
            io = EEGSignalIO(io_path=io_io_path, io_mode=io_mode)

            eeg = np.random.randn(32, 128)
            io_eeg_index = io.write_eeg(eeg)

            io_eeg = io.read_eeg(io_eeg_index)
            self.assertTrue(np.array_equal(eeg, io_eeg))

    def test_write_eeg_of_different_types(self):
        io_io_path = f'./tmp_out/{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'

        io_mode_space = ['pickle', 'lmdb']
        for io_mode in io_mode_space:
            io = EEGSignalIO(io_path=io_io_path, io_mode=io_mode)

            eeg = torch.randn(32, 128)
            io_eeg_index = io.write_eeg(eeg)

            eeg = np.random.randn(32, 128)
            io_eeg_index = io.write_eeg(eeg)

            io_eeg = io.read_eeg(io_eeg_index)
            self.assertTrue(np.array_equal(eeg, io_eeg))

    def test_write_eeg_of_different_shapes(self):
        io_io_path = f'./tmp_out/{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'

        io_mode_space = ['pickle', 'lmdb']
        for io_mode in io_mode_space:
            io = EEGSignalIO(io_path=io_io_path, io_mode=io_mode)

            eeg = np.random.randn(32, 128)
            io_eeg_index = io.write_eeg(eeg)

            eeg = np.random.randn(128, 9, 9)
            io_eeg_index = io.write_eeg(eeg)

            io_eeg = io.read_eeg(io_eeg_index)
            self.assertTrue(np.array_equal(eeg, io_eeg))


if __name__ == '__main__':
    unittest.main()


from typing import Callable, Dict, List

from ..base_transform import BaseTransform


class Lambda(BaseTransform):
    r'''
    Apply a user-defined lambda as a transform.

    .. code-block:: python

        transform = Lambda(targets=['y'], lambda x: x + 1)
        transform(y=1)['y']
        >>> 2

    Args:
        targets (list): What data to transform via the Lambda. (default: :obj:`['eeg', 'baseline', 'y']`)
        lambd (Callable): Lambda/function to be used for transform.

    .. automethod:: __call__
    '''
    def __init__(self,
                 lambd: Callable,
                 targets: List[str] = ['eeg', 'baseline', 'y']):
        super(Lambda, self).__init__()
        self._targets = targets
        self.lambd = lambd

    @property
    def targets(self) -> Dict[str, Callable]:
        return {target: self.apply for target in self._targets}

    def apply(self, *args, **kwargs) -> any:
        r'''
        Args:
            x (any): The input.

        Returns:
            any: The transformed output.
        '''
        return self.lambd(args[0])

    def __call__(self, *args, **kwargs) -> Dict[str, any]:
        r'''
        Args:
            x (any): The input.
        Returns:
            any: The transformed output.
        '''
        return super().__call__(*args, **kwargs)

    @property
    def repr_body(self) -> Dict:
        return dict(super().repr_body, **{
            'lambd': self.lambd,
            'targets': [...]
        })


from typing import Callable, Dict, List, Union

import numpy as np
import torch
from scipy.signal import hilbert
from torch_geometric.data import Data

from ..base_transform import EEGTransform


class ToG(EEGTransform):
    r'''
    A transformation method for constructing a graph representation of EEG signals, the results of which are applied to the input of the :obj:`torch_geometric` model. In the graph, nodes correspond to electrodes, and edges correspond to associations between electrodes (eg, spatially adjacent or functionally connected)

    :obj:`TorchEEG` provides some common graph structures. Consider using the following adjacency matrices depending on the dataset (with different EEG acquisition systems):

    - datasets.constants.emotion_recognition.deap.DEAP_ADJACENCY_MATRIX
    - datasets.constants.emotion_recognition.dreamer.DREAMER_ADJACENCY_MATRIX
    - datasets.constants.emotion_recognition.seed.SEED_ADJACENCY_MATRIX
    - ...

    .. code-block:: python

        transform = ToG(adj=DEAP_ADJACENCY_MATRIX)
        transform(eeg=np.random.randn(32, 128))['eeg']
        >>> torch_geometric.data.Data

    Args:
        adj (list): An adjacency matrix represented by a 2D array, each element in the adjacency matrix represents the electrode-to-electrode edge weight. Please keep the order of electrodes in the rows and columns of the adjacency matrix consistent with the EEG signal to be transformed.
        add_self_loop (bool): Whether to add self-loop edges to the graph. (default: :obj:`True`)
        threshold (float, optional): Used to cut edges when not None. Edges whose weights exceed a threshold are retained. (default: :obj:`None`)
        top_k (int, optional): Used to cut edges when not None. Keep the k edges connected to each node with the largest weights. (default: :obj:`None`)
        binary (bool): Whether to binarize the weights on the edges to 0 and 1. If set to True, binarization are done after topk and threshold, the edge weights that still have values are set to 1, otherwise they are set to 0. (default: :obj:`False`)
        complete_graph (bool): Whether to build as a complete graph. If False, only construct edges between electrodes based on non-zero elements; if True, construct variables between all electrodes and set the weight of non-existing edges to 0. (default: :obj:`False`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)
    
    .. automethod:: __call__
    '''
    def __init__(self,
                 adj: List[List],
                 add_self_loop: bool = True,
                 threshold: Union[float, None] = None,
                 top_k: Union[int, None] = None,
                 binary: bool = False,
                 complete_graph: bool = False,
                 apply_to_baseline: bool = False):
        super(ToG, self).__init__(apply_to_baseline=apply_to_baseline)

        self.add_self_loop = add_self_loop
        self.threshold = threshold
        self.top_k = top_k
        self.binary = binary
        self.complete_graph = complete_graph

        adj = torch.tensor(adj)

        if add_self_loop:
            adj = adj + torch.eye(adj.shape[0])

        if not self.threshold is None:
            adj[adj < self.threshold] = 0

        if not self.top_k is None:
            rows = []
            for row in adj:
                vals, index = row.topk(self.top_k)
                topk = torch.zeros_like(row)
                topk[index] = vals
                rows.append(topk)
            adj = torch.stack(rows)

        if self.binary:
            adj[adj != 0] = 1.0

        if self.complete_graph:
            adj[adj == 0] = 1e-6

        self.adj = adj.to_sparse()

    def __call__(self,
                 *args,
                 eeg: Union[np.ndarray, torch.Tensor],
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, Data]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch_geometric.data.Data: The graph representation data types that torch_geometric can accept. Nodes correspond to electrodes, and edges are determined via the given adjacency matrix.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def apply(self, eeg: Union[np.ndarray, torch.Tensor], **kwargs) -> Data:
        data = Data(edge_index=self.adj._indices())
        if isinstance(eeg, np.ndarray):
            data.x = torch.from_numpy(eeg).float()
        else:
            data.x = eeg
        data.edge_weight = self.adj._values()

        return data

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'adj': [...],
                'add_self_loop': self.add_self_loop,
                'threshold': self.threshold,
                'top_k': self.top_k,
                'binary': self.binary,
                'complete_graph': self.complete_graph
            })


class ToDynamicG(EEGTransform):
    r'''
    A transformation method for dynamically constructing the functional connections between electrodes according to the input EEG signals. The obtained graph structure based on functional connections can be applied to the input of the :obj:`torch_geometric` model. In the graph, nodes correspond to electrodes, and edges correspond to associations between electrodes (functionally connected)

    :obj:`TorchEEG` provides algorithms to dynamically calculate functional connections between electrodes:

    - Gaussian Distance
    - Absolute Pearson Correlation Coefficient
    - Phase Locking Value

    .. code-block:: python

        transform = ToDynamicG(edge_func='gaussian_distance', sigma=1.0, top_k=10, complete_graph=False)
        transform(eeg=np.random.randn(32, 128))['eeg']
        >>> Data(edge_index=[2, 320], x=[32, 128], edge_weight=[320])

        transform = ToDynamicG(edge_func='absolute_pearson_correlation_coefficient', threshold=0.1, binary=True)
        transform(eeg=np.random.randn(32, 128))['eeg']
        >>> Data(edge_index=[2, 310], x=[32, 128], edge_weight=[310])

        transform = ToDynamicG(edge_func='phase_locking_value')
        transform(eeg=np.random.randn(32, 128))['eeg']
        >>> Data(edge_index=[2, 992], x=[32, 128], edge_weight=[992])

        transform = ToDynamicG(edge_func=lambda x, y: (x * y).mean())
        transform(eeg=np.random.randn(32, 128))['eeg']
        >>> Data(edge_index=[2, 1024], x=[32, 128], edge_weight=[1024])

    Args:
        edge_func (str or Callable): Algorithms for computing functional connections. You can use the algorithms provided by TorchEEG, including gaussian_distance, absolute_pearson_correlation_coefficient and phase_locking_value. Or you can use custom functions by passing a callable object containing two parameters representing the signal of the two electrodes, and other named parameters (passed in when initializing the transform), and outputs the value of the functional connection between the two electrodes. (default: :obj:`gaussian_distance`)
        add_self_loop (bool): Whether to add self-loop edges to the graph. (default: :obj:`True`)
        threshold (float, optional): Used to cut edges when not None. Edges whose weights exceed a threshold are retained. (default: :obj:`None`)
        top_k (int, optional): Used to cut edges when not None. Keep the k edges connected to each node with the largest weights. (default: :obj:`None`)
        binary (bool): Whether to binarize the weights on the edges to 0 and 1. If set to True, binarization are done after topk and threshold, the edge weights that still have values are set to 1, otherwise they are set to 0. (default: :obj:`False`)
        complete_graph (bool): Whether to build as a complete graph. If False, only construct edges between electrodes based on non-zero elements; if True, construct variables between all electrodes and set the weight of non-existing edges to 0. (default: :obj:`False`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)
    
    .. automethod:: __call__
    '''
    def __init__(self,
                 edge_func: Union[str, Callable] = 'gaussian_distance',
                 add_self_loop: bool = True,
                 threshold: Union[float, None] = None,
                 top_k: Union[int, None] = None,
                 binary: bool = False,
                 complete_graph: bool = False,
                 apply_to_baseline: bool = False,
                 **kwargs):
        super(ToDynamicG, self).__init__(apply_to_baseline=apply_to_baseline)
        self.top_k = top_k
        self.threshold = threshold
        self.add_self_loop = add_self_loop
        self.binary = binary
        self.complete_graph = complete_graph
        self.edge_func = edge_func
        self.edge_func_list = {
            'gaussian_distance': self.gaussian_distance,
            'absolute_pearson_correlation_coefficient': self.absolute_pearson_correlation_coefficient,
            'phase_locking_value': self.phase_locking_value
        }
        self.kwargs = kwargs

    def opt(self, eeg: Union[np.ndarray, torch.Tensor]) -> np.ndarray:
        if isinstance(eeg, torch.Tensor):
            eeg = eeg.numpy()

        func = None

        if hasattr(self.edge_func, '__call__'):
            func = self.edge_func
        else:
            valid_edge_func = list(self.edge_func_list.keys())
            assert self.edge_func in valid_edge_func, f'Unimplemented calculation method {self.edge_func}, allowable selection of calculation methods include {valid_edge_func}'

            func = self.edge_func_list[self.edge_func]

        num_node = len(eeg)
        adj = np.zeros((num_node, num_node))

        for i in range(num_node):
            for j in range(num_node):
                adj[i][j] = func(eeg[i], eeg[j], **self.kwargs)

        return adj

    def adj(self, eeg: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        adj = torch.tensor(self.opt(eeg))

        if self.add_self_loop:
            adj = adj + torch.eye(adj.shape[0])

        if not self.threshold is None:
            adj[adj < self.threshold] = 0

        if not self.top_k is None:
            rows = []
            for row in adj:
                vals, index = row.topk(self.top_k)
                topk = torch.zeros_like(row)
                topk[index] = vals
                rows.append(topk)
            adj = torch.stack(rows)

        if self.binary:
            adj[adj != 0] = 1.0

        if self.complete_graph:
            adj[adj == 0] = 1e-6

        return adj.to_sparse()

    def gaussian_distance(self, x: np.ndarray, y: np.ndarray, sigma: float = 1.0, **kwargs):
        return np.exp(-np.linalg.norm(x - y, 2)**2 / (2. * sigma**2))

    def absolute_pearson_correlation_coefficient(self, x: np.ndarray, y: np.ndarray, **kwargs):
        return np.abs(np.corrcoef(x, y)[0][1])

    def phase_locking_value(self, x: np.ndarray, y: np.ndarray, **kwargs):
        x_hill = hilbert(x)
        y_hill = hilbert(y)
        pdt = (np.inner(x_hill, np.conj(y_hill)) /
               (np.sqrt(np.inner(x_hill, np.conj(x_hill)) * np.inner(y_hill, np.conj(y_hill)))))
        return np.angle(pdt)

    def __call__(self,
                 *args,
                 eeg: Union[np.ndarray, torch.Tensor],
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, Data]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (torch.Tensor, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.

        Returns:
            torch_geometric.data.Data: The graph representation data types that torch_geometric can accept. Nodes correspond to electrodes, and edges are determined via the given adjacency matrix.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def apply(self, eeg: Union[np.ndarray, torch.Tensor], **kwargs) -> Data:
        adj = self.adj(eeg)
        data = Data(edge_index=adj._indices())
        if isinstance(eeg, np.ndarray):
            data.x = torch.from_numpy(eeg).float()
        else:
            data.x = eeg
        data.edge_weight = adj._values()

        return data

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'top_k': self.top_k,
                'threshold': self.threshold,
                'add_self_loop': self.add_self_loop,
                'binary': self.binary,
                'complete_graph': self.complete_graph,
                'edge_func': self.edge_func
            })



import unittest
import torch
import numpy as np
from torcheeg.transforms import ToTensor, Resize, RandomNoise, RandomMask, RandomWindowSlice, RandomWindowWarp, RandomPCANoise, RandomFlip, RandomShift, RandomChannelShuffle, RandomFrequencyShift, RandomSignFlip, Contrastive, RandomMaskGrid, RandomMask2D


class TestTorchTransforms(unittest.TestCase):
    def test_contrastive(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = Contrastive(RandomNoise(), num_views=2)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'][0].shape), (32, 128))
        self.assertEqual(tuple(transformed_eeg['eeg'][1].shape), (32, 128))
        # transformed_eeg['eeg'][0] should be different from transformed_eeg['eeg'][1]
        self.assertFalse(torch.allclose(transformed_eeg['eeg'][0], transformed_eeg['eeg'][1]))

    def test_to_tensor(self):
        eeg = np.random.randn(32, 128)
        transformed_eeg = ToTensor()(eeg=eeg)
        self.assertEqual(eeg.shape, tuple(transformed_eeg['eeg'].shape))
        self.assertTrue(isinstance(transformed_eeg['eeg'], torch.Tensor))

    def test_resize(self):
        eeg = torch.randn(128, 9, 9)
        transformed_eeg = Resize(size=(64, 64))(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (128, 64, 64))

    def test_random_noise(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomNoise(p=1.0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (32, 128))

    def test_random_mask(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomMask(p=1.0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (32, 128))
    
    def test_random_window_slice(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomWindowSlice(p=1.0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (32, 128))

        eeg = torch.randn(1, 32, 128)
        transformed_eeg = RandomWindowSlice(p=1.0, window_size=100)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (1, 32, 128))

        eeg = torch.randn(128, 9, 9)
        transformed_eeg = RandomWindowSlice(p=1.0, series_dim=0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (128, 9, 9))
    
    def test_random_window_warp(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomWindowWarp(p=1.0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (32, 128))

        eeg = torch.randn(1, 32, 128)
        transformed_eeg = RandomWindowWarp(p=1.0, window_size=24, warp_size=48)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (1, 32, 128))

        eeg = torch.randn(128, 9, 9)
        transformed_eeg = RandomWindowWarp(p=1.0, series_dim=0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (128, 9, 9))

    def test_random_pca_noise(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomPCANoise(p=1.0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (32, 128))

        eeg = torch.randn(1, 32, 128)
        transformed_eeg = RandomPCANoise(p=1.0, mean=0.5, std=2.0, n_components=4)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (1, 32, 128))

        eeg = torch.randn(128, 9, 9)
        transformed_eeg = RandomPCANoise(p=1.0, series_dim=0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (128, 9, 9))
    
    def test_random_flip(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomFlip(p=1.0, dim=-1)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (32, 128))
    
    def test_random_sign_flip(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomSignFlip(p=1.0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (32, 128))

    def test_random_shift(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomShift(p=1.0, dim=-1, shift_min=8, shift_max=24)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (32, 128))

    def test_random_channel_shuffle(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomChannelShuffle(p=1.0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (32, 128))

    def test_random_sampling_rate_shift(self):
        eeg = torch.randn(32, 128)
        transformed_eeg = RandomFrequencyShift(p=1.0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (32, 128))

        eeg = torch.randn(1, 32, 128)
        transformed_eeg = RandomFrequencyShift(p=1.0, sampling_rate=128, shift_min=-1.0, shift_max=1.0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (1, 32, 128))

        eeg = torch.randn(128, 9, 9)
        transformed_eeg = RandomFrequencyShift(p=1.0, series_dim=0)(eeg=eeg)
        self.assertEqual(tuple(transformed_eeg['eeg'].shape), (128, 9, 9))

    def test_random_mask_2d(self):
        eeg = torch.randn(1, 32, 128)
        transformed_eeg = RandomMask2D()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (1, 32, 128))
    
    def test_random_mask_grid(self):
        eeg = torch.randn(32, 9, 9)
        transformed_eeg = RandomMaskGrid()(eeg=eeg)
        self.assertEqual(transformed_eeg['eeg'].shape, (32, 9, 9))


if __name__ == '__main__':
    unittest.main()


import logging
from itertools import chain
from typing import List, Tuple

import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn as nn
import torchmetrics
from torch.autograd import Function
from torch.utils.data import DataLoader

from ..classifier import ClassifierTrainer, classification_metrics

log = logging.getLogger(__name__)


class DualDataLoader:
    def __init__(self, ref_dataloader: DataLoader,
                 other_dataloader: DataLoader):
        self.ref_dataloader = ref_dataloader
        self.other_dataloader = other_dataloader

    def __iter__(self):
        return self.dual_iterator()

    def __len__(self):
        return len(self.ref_dataloader)

    def dual_iterator(self):
        other_it = iter(self.other_dataloader)
        for data in self.ref_dataloader:
            try:
                data_ = next(other_it)
            except StopIteration:
                other_it = iter(self.other_dataloader)
                data_ = next(other_it)
            yield data, data_


class GradientReverse(Function):
    @staticmethod
    def forward(ctx, x, alpha):
        ctx.alpha = alpha
        return x.view_as(x)

    @staticmethod
    def backward(ctx, grad_output):
        output = grad_output.neg() * ctx.alpha
        return output, None


class _DANNLikeTrainer(ClassifierTrainer):
    def __init__(self,
                 extractor: nn.Module,
                 classifier: nn.Module,
                 domain_classifier: nn.Module,
                 num_classes: int,
                 lr: float = 1e-4,
                 weight_decay: float = 0.0,
                 weight_domain: float = 1.0,
                 alpha_scheduler: bool = True,
                 lr_scheduler_gamma: float = 0.0,
                 lr_scheduler_decay: float = 0.75,
                 warmup_epochs: int = 0,
                 devices: int = 1,
                 accelerator: str = "cpu",
                 metrics: List[str] = ["accuracy"]):
        super(ClassifierTrainer, self).__init__()

        self.extractor = extractor
        self.classifier = classifier
        self.domain_classifier = domain_classifier

        self.lr = lr
        self.weight_decay = weight_decay
        self.weight_domain = weight_domain
        self.alpha_scheduler = alpha_scheduler
        self.lr_scheduler = not lr_scheduler_gamma == 0.0
        self.lr_scheduler_gamma = lr_scheduler_gamma
        self.lr_scheduler_decay = lr_scheduler_decay
        self.warmup_epochs = warmup_epochs

        self.devices = devices
        self.accelerator = accelerator
        self.metrics = metrics

        self.num_classes = num_classes
        self.metrics = metrics
        self.init_metrics(metrics, num_classes)

        self.ce_fn = nn.CrossEntropyLoss()

        self.num_batches = None  # init in 'fit' method
        self.non_warmup_epochs = None  # init in 'fit' method
        self.lr_factor = 1.0  # update in 'on_train_batch_start' method
        self.scheduled_alpha = 1.0  # update in 'on_train_batch_start' method

    def init_metrics(self, metrics: List[str], num_classes: int) -> None:
        self.train_domain_loss = torchmetrics.MeanMetric()
        self.train_task_loss = torchmetrics.MeanMetric()

        self.val_loss = torchmetrics.MeanMetric()
        self.test_loss = torchmetrics.MeanMetric()

        self.train_metrics = classification_metrics(metrics, num_classes)
        self.val_metrics = classification_metrics(metrics, num_classes)
        self.test_metrics = classification_metrics(metrics, num_classes)

    def fit(self,
            source_loader: DataLoader,
            target_loader: DataLoader,
            val_loader: DataLoader,
            max_epochs: int = 300,
            *args,
            **kwargs):
        r'''
        Args:
            source_loader (DataLoader): Iterable DataLoader for traversing the data batch from the source domain (:obj:`torch.utils.data.dataloader.DataLoader`, :obj:`torch_geometric.loader.DataLoader`, etc).
            target_loader (DataLoader): Iterable DataLoader for traversing the training data batch from the target domain (:obj:`torch.utils.data.dataloader.DataLoader`, :obj:`torch_geometric.loader.DataLoader`, etc). The target dataset does not have to return labels.
            val_loader (DataLoader): Iterable DataLoader for traversing the validation data batch (:obj:`torch.utils.data.dataloader.DataLoader`, :obj:`torch_geometric.loader.DataLoader`, etc).
            max_epochs (int): The maximum number of epochs to train. (default: :obj:`300`)
        '''

        train_loader = DualDataLoader(source_loader, target_loader)

        self.num_batches = len(train_loader)
        self.non_warmup_epochs = max_epochs - self.warmup_epochs

        trainer = pl.Trainer(devices=self.devices,
                             accelerator=self.accelerator,
                             max_epochs=max_epochs,
                             *args,
                             **kwargs)
        return trainer.fit(self, train_loader, val_loader)

    def on_train_batch_start(self, batch, batch_idx):
        if self.current_epoch >= self.warmup_epochs:
            delta_epoch = self.current_epoch - self.warmup_epochs
            p = (batch_idx + delta_epoch * self.num_batches) / (
                self.non_warmup_epochs * self.num_batches)

            if self.alpha_scheduler:
                self.scheduled_alpha = 2.0 / (1.0 + np.exp(-10 * p)) - 1

            if self.lr_scheduler:
                self.lr_factor = 1.0 / ((1.0 + self.lr_scheduler_gamma * p)**
                                        self.lr_scheduler_decay)

    def _domain_loss_fn(self, x_source_feat: torch.Tensor,
                        x_target_feat: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError

    def training_step(self, batch: Tuple[torch.Tensor],
                      batch_idx: int) -> torch.Tensor:
        (x_source, y_source), (x_target, _) = batch

        x_source_feat = self.extractor(x_source)
        y_source_pred = self.classifier(x_source_feat)
        x_target_feat = self.extractor(x_target)

        x_source_feat_r = GradientReverse.apply(x_source_feat,
                                                self.scheduled_alpha)
        y_source_disc = self.domain_classifier(x_source_feat_r)

        x_target_feat_r = GradientReverse.apply(x_target_feat,
                                                self.scheduled_alpha)
        y_target_disc = self.domain_classifier(x_target_feat_r)

        task_loss = self.ce_fn(y_source_pred, y_source)
        domain_loss = self.ce_fn(
            y_source_disc,
            torch.zeros(len(y_source_disc),
                        dtype=torch.long,
                        device=x_source.device)) + self.ce_fn(
                            y_target_disc,
                            torch.ones(len(y_target_disc),
                                       dtype=torch.long,
                                       device=x_target.device))

        if self.current_epoch >= self.warmup_epochs:
            loss = task_loss + self.weight_domain * domain_loss
        else:
            loss = task_loss

        self.log("train_domain_loss",
                 self.train_domain_loss(loss),
                 prog_bar=True,
                 on_epoch=False,
                 logger=False,
                 on_step=True)
        self.log("train_task_loss",
                 self.train_task_loss(loss),
                 prog_bar=True,
                 on_epoch=False,
                 logger=False,
                 on_step=True)
        for i, metric_value in enumerate(self.train_metrics.values()):
            self.log(f"train_{self.metrics[i]}",
                     metric_value(y_source_pred, y_source),
                     prog_bar=True,
                     on_epoch=False,
                     logger=False,
                     on_step=True)

        return loss

    def on_train_epoch_end(self) -> None:
        self.log("train_domain_loss",
                 self.train_domain_loss.compute(),
                 prog_bar=False,
                 on_epoch=True,
                 on_step=False,
                 logger=True)
        self.log("train_task_loss",
                 self.train_task_loss.compute(),
                 prog_bar=False,
                 on_epoch=True,
                 on_step=False,
                 logger=True)
        for i, metric_value in enumerate(self.train_metrics.values()):
            self.log(f"train_{self.metrics[i]}",
                     metric_value.compute(),
                     prog_bar=False,
                     on_epoch=True,
                     on_step=False,
                     logger=True)

        # print the metrics
        str = "\n[Train] "
        for key, value in self.trainer.logged_metrics.items():
            if key.startswith("train_"):
                str += f"{key}: {value:.3f} "
        print(str + '\n')

        # reset the metrics
        self.train_domain_loss.reset()
        self.train_task_loss.reset()
        self.train_metrics.reset()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.extractor(x)
        x = self.classifier(x)
        return x

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(chain(self.extractor.parameters(),
                                           self.classifier.parameters(),
                                           self.domain_classifier.parameters()),
                                     lr=self.lr,
                                     weight_decay=self.weight_decay)
        if self.lr_scheduler:
            scheduler = torch.optim.lr_scheduler.LambdaLR(
                optimizer, lr_lambda=lambda epoch: self.lr_factor)
            return [optimizer], [scheduler]
        return [optimizer]


import os
import logging
import re
from typing import Any, Callable, Dict, List, Tuple, Union

import scipy.io as scio

from ...constants.emotion_recognition.amigos import (
    AMIGOS_ADJACENCY_MATRIX, AMIGOS_CHANNEL_LOCATION_DICT)
from ..base_dataset import BaseDataset

log = logging.getLogger(__name__)

class AMIGOSDataset(BaseDataset):
    r'''
    A dataset for Multimodal research of affect, personality traits and mood on Individuals and GrOupS (AMIGOS). This class generates training samples and test samples according to the given parameters, and caches the generated results in a unified input and output format (IO). The relevant information of the dataset is as follows:

    - Author: Miranda-Correa et al.
    - Year: 2018
    - Download URL: http://www.eecs.qmul.ac.uk/mmv/datasets/amigos/download.html
    - Reference: Miranda-Correa J A, Abadi M K, Sebe N, et al. Amigos: A dataset for affect, personality and mood research on individuals and groups[J]. IEEE Transactions on Affective Computing, 2018, 12(2): 479-493.
    - Stimulus: 16 short affective video extracts and 4 long affective video extracts from movies.
    - Signals: Electroencephalogram (14 channels at 128Hz), electrocardiogram (2 channels at 60Hz) and galvanic skin response (1 channel at 60Hz) of 40 subjects. For the first 16 trials, 40 subjects watched a set of short affective video extracts. For the last 4 trials, 37 of the participants of the previous experiment watched a set of long affective video extracts.
    - Rating: arousal (1-9), valence (1-9), dominance (1-9), liking (1-9), familiarity (1-9), neutral (0, 1), disgust (0, 1),happiness (0, 1), surprise (0, 1), anger (0, 1), fear (0, 1), and sadness (0, 1).

    In order to use this dataset, the download folder :obj:`data_preprocessed` is required, containing the following files:
    
    - Data_Preprocessed_P01.mat
    - Data_Preprocessed_P02.mat
    - Data_Preprocessed_P03.mat
    - ...
    - Data_Preprocessed_P40.mat

    An example dataset for CNN-based methods:

    .. code-block:: python

        dataset = AMIGOSDataset(io_path=f'./amigos',
                                root_path='./data_preprocessed',
                                offline_transform=transforms.Compose([
                                    transforms.BandDifferentialEntropy(),
                                    transforms.ToGrid(AMIGOS_CHANNEL_LOCATION_DICT)
                                ]),
                                online_transform=transforms.ToTensor(),
                                label_transform=transforms.Compose([
                                    transforms.Select('valence'),
                                    transforms.Binary(5.0),
                                ]))
        print(dataset[0])
        # EEG signal (torch.Tensor[4, 9, 9]),
        # coresponding baseline signal (torch.Tensor[4, 9, 9]),
        # label (int)

    Another example dataset for CNN-based methods:

    .. code-block:: python

        dataset = AMIGOSDataset(io_path=f'./amigos',
                              root_path='./data_preprocessed',
                              online_transform=transforms.Compose([
                                  transforms.To2d(),
                                  transforms.ToTensor()
                              ]),
                              label_transform=transforms.Compose([
                                  transforms.Select('valence'),
                                  transforms.Binary(5.0),
                              ]))
        print(dataset[0])
        # EEG signal (torch.Tensor[14, 128]),
        # coresponding baseline signal (torch.Tensor[14, 128]),
        # label (int)

    An example dataset for GNN-based methods:

    .. code-block:: python
    
        dataset = AMIGOSDataset(io_path=f'./amigos',
                              root_path='./data_preprocessed',
                              online_transform=transforms.Compose([
                                  ToG(AMIGOS_ADJACENCY_MATRIX)
                              ]),
                              label_transform=transforms.Compose([
                                  transforms.Select('valence'),
                                  transforms.Binary(5.0),
                              ]))
        print(dataset[0])
        # EEG signal (torch_geometric.data.Data),
        # coresponding baseline signal (torch_geometric.data.Data),
        # label (int)
    
    In particular, TorchEEG utilizes the producer-consumer model to allow multi-process data preprocessing. If your data preprocessing is time consuming, consider increasing :obj:`num_worker` for higher speedup. If running under Windows, please use the proper idiom in the main module:

    .. code-block:: python
    
        if __name__ == '__main__':
            dataset = AMIGOSDataset(io_path=f'./amigos',
                                    root_path='./data_preprocessed',
                                    online_transform=transforms.Compose([
                                        ToG(AMIGOS_ADJACENCY_MATRIX)
                                    ]),
                                    label_transform=transforms.Compose([
                                        transforms.Select('valence'),
                                        transforms.Binary(5.0),
                                    ]),
                                    num_worker=4)
            print(dataset[0])
            # EEG signal (torch_geometric.data.Data),
            # coresponding baseline signal (torch_geometric.data.Data),
            # label (int)

    Args:
        root_path (str): Downloaded data files in matlab (unzipped data_preprocessed.zip) formats (default: :obj:`'./data_preprocessed'`)
        chunk_size (int): Number of data points included in each EEG chunk as training or test samples. If set to -1, the EEG signal of a trial is used as a sample of a chunk. (default: :obj:`128`)
        overlap (int): The number of overlapping data points between different chunks when dividing EEG chunks. (default: :obj:`0`)
        num_channel (int): Number of channels used, of which the first 14 channels are EEG signals. (default: :obj:`14`)
        num_trial (int): Number of trials used, of which the first 16 trials are conducted with short videos and the last 4 trials are conducted with long videos. If set to -1, all trials are used. (default: :obj:`16`)
        skipped_subjects (int): The participant ID to be removed because there are some invalid data in the preprocessed version. (default: :obj:`[9, 12, 21, 22, 23, 24, 33]`)
        num_baseline (int): Number of baseline signal chunks used. (default: :obj:`5`)
        baseline_chunk_size (int): Number of data points included in each baseline signal chunk. The baseline signal in the AMIGOS dataset has a total of 640 data points. (default: :obj:`128`)
        online_transform (Callable, optional): The transformation of the EEG signals and baseline EEG signals. The input is a :obj:`np.ndarray`, and the ouput is used as the first and second value of each element in the dataset. (default: :obj:`None`)
        offline_transform (Callable, optional): The usage is the same as :obj:`online_transform`, but executed before generating IO intermediate results. (default: :obj:`None`)
        label_transform (Callable, optional): The transformation of the label. The input is an information dictionary, and the ouput is used as the third value of each element in the dataset. (default: :obj:`None`)
        before_trial (Callable, optional): The hook performed on the trial to which the sample belongs. It is performed before the offline transformation and thus typically used to implement context-dependent sample transformations, such as moving averages, etc. The input of this hook function is a 2D EEG signal with shape (number of electrodes, number of data points), whose ideal output shape is also (number of electrodes, number of data points).
        after_trial (Callable, optional): The hook performed on the trial to which the sample belongs. It is performed after the offline transformation and thus typically used to implement context-dependent sample transformations, such as moving averages, etc. The input and output of this hook function should be a sequence of dictionaries representing a sequence of EEG samples. Each dictionary contains two key-value pairs, indexed by :obj:`eeg` (the EEG signal matrix) and :obj:`key` (the index in the database) respectively.
        io_path (str): The path to generated unified data IO, cached as an intermediate result. (default: :obj:`./io/amigos`)
        io_size (int): Maximum size database may grow to; used to size the memory mapping. If database grows larger than ``map_size``, an exception will be raised and the user must close and reopen. (default: :obj:`10485760`)
        io_mode (str): Storage mode of EEG signal. When io_mode is set to :obj:`lmdb`, TorchEEG provides an efficient database (LMDB) for storing EEG signals. LMDB may not perform well on limited operating systems, where a file system based EEG signal storage is also provided. When io_mode is set to :obj:`pickle`, pickle-based persistence files are used. (default: :obj:`lmdb`)
        num_worker (int): Number of subprocesses to use for data loading. 0 means that the data will be loaded in the main process. (default: :obj:`0`)
        verbose (bool): Whether to display logs during processing, such as progress bars, etc. (default: :obj:`True`)
        in_memory (bool): Whether to load the entire dataset into memory. If :obj:`in_memory` is set to True, then the first time an EEG sample is read, the entire dataset is loaded into memory for subsequent retrieval. Otherwise, the dataset is stored on disk to avoid the out-of-memory problem. (default: :obj:`False`)
    '''
    channel_location_dict = AMIGOS_CHANNEL_LOCATION_DICT
    adjacency_matrix = AMIGOS_ADJACENCY_MATRIX

    def __init__(self,
                 root_path: str = './data_preprocessed',
                 chunk_size: int = 128,
                 overlap: int = 0,
                 num_channel: int = 14,
                 num_trial: int = 16,
                 skipped_subjects: List[int] = [9, 12, 21, 22, 23, 24, 33],
                 num_baseline: int = 5,
                 baseline_chunk_size: int = 128,
                 online_transform: Union[None, Callable] = None,
                 offline_transform: Union[None, Callable] = None,
                 label_transform: Union[None, Callable] = None,
                 before_trial: Union[None, Callable] = None,
                 after_trial: Union[Callable, None] = None,
                 after_session: Union[Callable, None] = None,
                 after_subject: Union[Callable, None] = None,
                 io_path: str = './io/amigos',
                 io_size: int = 10485760,
                 io_mode: str = 'lmdb',
                 num_worker: int = 0,
                 verbose: bool = True,
                 in_memory: bool = False):
        # pass all arguments to super class
        params = {
            'root_path': root_path,
            'chunk_size': chunk_size,
            'overlap': overlap,
            'num_channel': num_channel,
            'num_trial': num_trial,
            'skipped_subjects': skipped_subjects,
            'num_baseline': num_baseline,
            'baseline_chunk_size': baseline_chunk_size,
            'online_transform': online_transform,
            'offline_transform': offline_transform,
            'label_transform': label_transform,
            'before_trial': before_trial,
            'after_trial': after_trial,
            'after_session': after_session,
            'after_subject': after_subject,
            'io_path': io_path,
            'io_size': io_size,
            'io_mode': io_mode,
            'num_worker': num_worker,
            'verbose': verbose,
            'in_memory': in_memory
        }
        super().__init__(**params)
        # save all arguments to __dict__
        self.__dict__.update(params)

    @staticmethod
    def process_record(file: Any = None,
                   root_path: str = './data_preprocessed',
                   chunk_size: int = 128,
                   overlap: int = 0,
                   num_channel: int = 14,
                   num_trial: int = 16,
                   skipped_subjects: List[int] = [9, 12, 21, 22, 23, 24, 33],
                   num_baseline: int = 5,
                   baseline_chunk_size: int = 128,
                   before_trial: Union[None, Callable] = None,
                   offline_transform: Union[None, Callable] = None,
                   after_trial: Union[None, Callable] = None,
                   **kwargs):
        file_name = file  # an element from file name list

        subject = int(
            re.findall(r'Data_Preprocessed_P(\d*).mat',
                       file_name)[0])  # subject (40)

        if subject in skipped_subjects:
            return

        data = scio.loadmat(os.path.join(root_path, file_name),
                            verify_compressed_data_integrity=False)
        samples = data['joined_data'][
            0]  # trial (20), timestep(n*128), channel(17) (14 channels are EEGs)
        # label file
        labels = data['labels_selfassessment'][
            0]  # trial (20), label of different dimensions ((1, 12))

        write_pointer = 0

        max_len = len(samples)
        if not (num_trial <= 0):
            max_len = min(len(samples), num_trial)
            
        # loop for each trial
        for trial_id in range(max_len):
            
            # extract baseline signals
            trial_samples = samples[trial_id]

            # record the common meta info
            trial_meta_info = {'subject_id': subject, 'trial_id': trial_id}
            trial_rating = labels[trial_id][
                0]  # label of different dimensions (12)

            # missing values
            if (not sum(trial_samples.shape)) or (not sum(trial_rating.shape)):
                # 3 of the participants (08,24,28<->32) of the previous experiment did not watch a set of 4 long affective
                if sum(trial_samples.shape) != sum(trial_rating.shape):
                    print(
                        f'Find EEG signals without labels, or labels without EEG signals. Please check the {trial_id + 1}-th experiment of the {subject}-th subject in the file {file_name}. TorchEEG currently skipped the mismatched data.'
                    )
                continue

            trial_samples = trial_samples.swapaxes(
                1, 0)  # channel(17), timestep(n*128)
            if before_trial:
                trial_samples = before_trial(trial_samples)

            for label_index, label_name in enumerate([
                    'arousal', 'valence', 'dominance', 'liking', 'familiarity',
                    'neutral', 'disgust', 'happiness', 'surprise', 'anger',
                    'fear', 'sadness'
            ]):
                trial_meta_info[label_name] = trial_rating[label_index]

            # extract baseline signals
            trial_baseline_sample = trial_samples[:num_channel, :
                                                  baseline_chunk_size *
                                                  num_baseline]  # channel(14), timestep(5*128)
            trial_baseline_sample = trial_baseline_sample.reshape(
                num_channel, num_baseline,
                baseline_chunk_size).mean(axis=1)  # channel(14), timestep(128)

            start_at = baseline_chunk_size * num_baseline
            if chunk_size <= 0:
                dynamic_chunk_size = trial_samples.shape[1] - start_at
            else:
                dynamic_chunk_size = chunk_size

            # chunk with chunk size
            end_at = start_at + dynamic_chunk_size
            # calculate moving step
            step = dynamic_chunk_size - overlap

            while end_at <= trial_samples.shape[1]:
                clip_sample = trial_samples[:num_channel, start_at:end_at]

                t_eeg = clip_sample
                t_baseline = trial_baseline_sample
                if not offline_transform is None:
                    t = offline_transform(eeg=clip_sample,
                                          baseline=trial_baseline_sample)
                    t_eeg = t['eeg']
                    t_baseline = t['baseline']

                # put baseline signal into IO
                if not 'baseline_id' in trial_meta_info:
                    trial_base_id = f'{file_name}_{write_pointer}'
                    yield {'eeg': t_baseline, 'key': trial_base_id}
                    write_pointer += 1
                    trial_meta_info['baseline_id'] = trial_base_id

                clip_id = f'{file_name}_{write_pointer}'
                write_pointer += 1

                # record meta info for each signal
                record_info = {
                    'start_at': start_at,
                    'end_at': end_at,
                    'clip_id': clip_id
                }
                record_info.update(trial_meta_info)
                yield {'eeg': t_eeg, 'key': clip_id, 'info': record_info}

                start_at = start_at + step
                end_at = start_at + dynamic_chunk_size

    def set_records(self, root_path: str = './data_preprocessed', **kwargs):
        return os.listdir(root_path)

    def __getitem__(self, index: int) -> Tuple[any, any, int, int, int]:
        info = self.read_info(index)

        eeg_index = str(info['clip_id'])
        eeg_record = str(info['_record_id'])
        eeg = self.read_eeg(eeg_record, eeg_index)

        baseline_index = str(info['baseline_id'])
        baseline = self.read_eeg(eeg_record, baseline_index)

        signal = eeg
        label = info

        if self.online_transform:
            signal = self.online_transform(eeg=eeg, baseline=baseline)['eeg']

        if self.label_transform:
            label = self.label_transform(y=info)['y']

        return signal, label

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'root_path': self.root_path,
                'chunk_size': self.chunk_size,
                'overlap': self.overlap,
                'num_channel': self.num_channel,
                'num_trial': self.num_trial,
                'skipped_subjects': self.skipped_subjects,
                'num_baseline': self.num_baseline,
                'baseline_chunk_size': self.baseline_chunk_size,
                'online_transform': self.online_transform,
                'offline_transform': self.offline_transform,
                'label_transform': self.label_transform,
                'before_trial': self.before_trial,
                'after_trial': self.after_trial,
                'num_worker': self.num_worker,
                'verbose': self.verbose,
                'io_size': self.io_size
            })



import logging
import os
import shutil
from typing import Any, Callable, Dict
import torch
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from torch.utils.data import Dataset
from tqdm import tqdm

from torcheeg.io import EEGSignalIO, MetaInfoIO

MAX_QUEUE_SIZE = 1024

log = logging.getLogger(__name__)


class MockLock():
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class BaseDataset(Dataset):
    def __init__(self,
                 io_path: str = None,
                 io_size: int = 10485760,
                 io_mode: str = 'lmdb',
                 in_memory: bool = False,
                 num_worker: int = 0,
                 verbose: bool = True,
                 after_trial: Callable = None,
                 after_session: Callable = None,
                 after_subject: Callable = None,
                 **kwargs):
        self.io_path = io_path
        self.io_size = io_size
        self.io_mode = io_mode
        self.in_memory = in_memory
        self.num_worker = num_worker
        self.verbose = verbose
        self.after_trial = after_trial
        self.after_session = after_session
        self.after_subject = after_subject

        # new IO
        if not self.exist(self.io_path):
            print(
                f'dataset does not exist at path {self.io_path}, generating files to path...'
            )
            # make the root dictionary
            os.makedirs(self.io_path, exist_ok=True)

            records = self.set_records(**kwargs)
            if self.num_worker == 0:
                try:
                    for file_id, file in tqdm(enumerate(records),
                                              disable=not self.verbose,
                                              desc="[PROCESS]",
                                              total=len(records)):
                        self.save_record(io_path=self.io_path,
                                         io_size=self.io_size,
                                         io_mode=self.io_mode,
                                         file=file,
                                         file_id=file_id,
                                         process_record=self.process_record,
                                         **kwargs)
                except Exception as e:
                    # shutil to delete the database
                    shutil.rmtree(self.io_path)
                    raise e
            else:
                # catch the exception
                try:
                    Parallel(n_jobs=self.num_worker)(
                        delayed(self.save_record)(
                            io_path=io_path,
                            io_size=io_size,
                            io_mode=io_mode,
                            file_id=file_id,
                            file=file,
                            process_record=self.process_record,
                            **kwargs)
                        for file_id, file in tqdm(enumerate(records),
                                                  disable=not self.verbose,
                                                  desc="[PROCESS]",
                                                  total=len(records)))
                except Exception as e:
                    # shutil to delete the database
                    shutil.rmtree(self.io_path)
                    raise e

            # get the global io
            self.eeg_io_router, self.info = self.get_pointer(io_path=io_path,
                                                             io_size=io_size,
                                                             io_mode=io_mode)
            # catch the exception
            try:
                self.post_process_record(after_trial=after_trial,
                                         after_session=after_session,
                                         after_subject=after_subject)
            except Exception as e:
                # shutil to delete the database
                shutil.rmtree(self.io_path)
                raise e
        else:
            print(
                f'dataset already exists at path {self.io_path}, reading from path...'
            )
            # get the global io
            self.eeg_io_router, self.info = self.get_pointer(io_path=io_path,
                                                             io_size=io_size,
                                                             io_mode=io_mode)

    def set_records(self, **kwargs):
        '''
        The block method for generating the database. It is used to describe which data blocks need to be processed to generate the database. It is called in parallel by :obj:`joblib.Parallel` in :obj:`__init__` of the class.

        Args:
            lock (joblib.parallel.Lock): The lock for IO writter. (default: :obj:`None`)
            **kwargs: The arguments derived from __init__ of the class.

        .. code-block:: python

        def set_records(self, root_path: str = None, **kwargs):
                # e.g., return file name list for process_record to process
                return os.listdir(root_path)

        '''
        raise NotImplementedError(
            "Method set_records is not implemented in class BaseDataset")

    def get_pointer(self, io_path: str, io_size: int, io_mode: str):
        '''
        The method for initializing the database. It is used to describe how to initialize the database. It is called in :obj:`__init__` of the class.
        '''
        # get all records
        records = os.listdir(io_path)
        # filter the records with the prefix '_record_'
        records = list(filter(lambda x: '_record_' in x, records))

        # for every record, get the io_path, and init the info_io and eeg_io
        eeg_io_router = {}
        info_merged = None

        for record in records:
            meta_info_io_path = os.path.join(io_path, record, 'info.csv')
            eeg_signal_io_path = os.path.join(io_path, record, 'eeg')
            info_io = MetaInfoIO(meta_info_io_path)
            eeg_io = EEGSignalIO(eeg_signal_io_path,
                                 io_size=io_size,
                                 io_mode=io_mode)
            eeg_io_router[record] = eeg_io
            info_df = info_io.read_all()

            assert '_record_id' not in info_df.columns, \
                "column '_record_id' is a forbidden reserved word and is used to index the corresponding IO. Please replace your '_record_id' with another name."
            info_df['_record_id'] = record

            if info_merged is None:
                info_merged = info_df
            else:
                info_merged = pd.concat([info_merged, info_df],
                                        ignore_index=True)
        return eeg_io_router, info_merged

    @staticmethod
    def save_record(io_path: str = None,
                    io_size: int = 10485760,
                    io_mode: str = 'lmdb',
                    file: Any = None,
                    file_id: int = None,
                    process_record=None,
                    **kwargs):
        _record_id = str(file_id)
        meta_info_io_path = os.path.join(io_path, f'_record_{_record_id}',
                                         'info.csv')
        eeg_signal_io_path = os.path.join(io_path, f'_record_{_record_id}',
                                          'eeg')

        info_io = MetaInfoIO(meta_info_io_path)
        eeg_io = EEGSignalIO(eeg_signal_io_path,
                             io_size=io_size,
                             io_mode=io_mode)

        gen = process_record(file=file, **kwargs)
        # loop for data yield by process_record, until to the end of the data
        while True:
            try:
                # call process_record of the class
                # get the current class name
                obj = next(gen)

            except StopIteration:
                break

            if 'eeg' in obj and 'key' in obj:
                eeg_io.write_eeg(obj['eeg'], obj['key'])
            if 'info' in obj:
                info_io.write_info(obj['info'])

    @staticmethod
    def process_record(file: Any = None, **kwargs):
        '''
        The IO method for generating the database. It is used to describe how files are processed to generate the database. It is called in parallel by :obj:`joblib.Parallel` in :obj:`__init__` of the class.

        Args:
            file (Any): The file to be processed. It is an element in the list returned by set_records. (default: :obj:`Any`)
            **kwargs: The arguments derived from :obj:`__init__` of the class.

        .. code-block:: python

            def process_record(file: Any = None, chunk_size: int = 128, **kwargs):
                # process file
                eeg = np.ndarray((chunk_size, 64, 128), dtype=np.float32)
                key = '1'
                info = {
                    'subject': '1',
                    'session': '1',
                    'run': '1',
                    'label': '1'
                }
                yield {
                    'eeg': eeg,
                    'key': key,
                    'info': info
                }

        '''

        raise NotImplementedError(
            "Method process_record is not implemented in class BaseDataset")

    def post_process_record(self,
                            after_trial: Callable = None,
                            after_session: Callable = None,
                            after_subject: Callable = None):
        '''
        The hook method for post-processing the data. It is used to describe how to post-process the data.
        '''
        pbar = tqdm(total=len(self),
                    disable=not self.verbose,
                    desc="[POST-PROCESS]")

        # if all the hooks are None, then return
        if after_trial is None and after_session is None and after_subject is None:
            return

        if 'subject_id' in self.info.columns:
            subject_df = self.info.groupby('subject_id')
        else:
            subject_df = [(None, self.info)]
            if not after_subject is None:
                print(
                    "No subject_id column found in info, after_subject hook is ignored."
                )
        if after_subject is None:
            after_subject = lambda x: x

        for _, subject_info in subject_df:

            subject_record_list = []
            subject_index_list = []
            subject_samples = []

            # check if have a session_id column
            if 'session_id' in subject_info.columns:
                session_df = subject_info.groupby('session_id')
            else:
                session_df = [(None, subject_info)]
                if not after_session is None:
                    print(
                        "No session_id column found in info, after_session hook is ignored."
                    )
            if after_session is None:
                after_session = lambda x: x

            for _, session_info in session_df:

                # check if have a trial_id column
                if 'trial_id' in session_info.columns:
                    trial_df = session_info.groupby('trial_id')
                else:
                    trial_df = [(None, session_info)]
                    if not after_trial is None:
                        print(
                            "No trial_id column found in info, after_trial hook is ignored."
                        )
                if after_trial is None:
                    after_trial = lambda x: x

                session_samples = []
                for _, trial_info in trial_df:
                    trial_samples = []
                    for i in range(len(trial_info)):
                        eeg_index = str(trial_info.iloc[i]['clip_id'])
                        eeg_record = str(trial_info.iloc[i]['_record_id'])

                        # if i == 0 and 'baseline_id' in trial_info.columns:
                        #     baseline_index = str(
                        #         trial_info.iloc[i]['baseline_id'])
                        #     subject_record_list.append(eeg_record)
                        #     subject_index_list.append(baseline_index)

                        #     eeg = self.read_eeg(eeg_record, baseline_index)
                        #     trial_samples += [eeg]

                        subject_record_list.append(eeg_record)
                        subject_index_list.append(eeg_index)

                        eeg = self.read_eeg(eeg_record, eeg_index)
                        trial_samples += [eeg]

                        pbar.update(1)

                    trial_samples = self.hook_data_interface(
                        after_trial, trial_samples)
                    session_samples += trial_samples

                session_samples = self.hook_data_interface(
                    after_session, session_samples)
                subject_samples += session_samples

            subject_samples = self.hook_data_interface(after_subject,
                                                       subject_samples)

            # save the data
            for i, eeg in enumerate(subject_samples):
                eeg_index = str(subject_index_list[i])
                eeg_record = str(subject_record_list[i])

                self.write_eeg(eeg_record, eeg_index, eeg)

        pbar.close()

    @staticmethod
    def hook_data_interface(hook: Callable, data: Any):
        # like [np.random.randn(32, 128), np.random.randn(32, 128)]
        if isinstance(data[0], np.ndarray):
            data = np.stack(data, axis=0)
        elif isinstance(data[0], torch.Tensor):
            data = torch.stack(data, axis=0)
        # else list

        # shape like (2, 32, 128)
        data = hook(data)

        # back to list like [np.random.randn(32, 128), np.random.randn(32, 128)]
        if isinstance(data, np.ndarray):
            data = np.split(data, data.shape[0], axis=0)
            data = [np.squeeze(d, axis=0) for d in data]
        elif isinstance(data, torch.Tensor):
            data = torch.split(data, data.shape[0], dim=0)
            data = [torch.squeeze(d, axis=0) for d in data]
        # else list
        return data

    def read_eeg(self, record: str, key: str) -> Any:
        r'''
        Query the corresponding EEG signal in the EEGSignalIO according to the the given :obj:`key`. If :obj:`self.in_memory` is set to :obj:`True`, then EEGSignalIO will be read into memory and directly index the specified EEG signal in memory with the given :obj:`key` on subsequent reads.

        Args:
            record (str): The record id of the EEG signal to be queried.
            key (str): The index of the EEG signal to be queried.
            
        Returns:
            any: The EEG signal sample.
        '''
        eeg_io = self.eeg_io_router[record]
        if self.in_memory:
            return eeg_io.read_eeg_in_memory(key)
        return eeg_io.read_eeg(key)

    def write_eeg(self, record: str, key: str, eeg: Any):
        r'''
        Update the corresponding EEG signal in the EEGSignalIO according to the the given :obj:`key`.

        Args:
            record (str): The record id of the EEG signal to be queried.
            key (str): The index of the EEG signal to be queried.
            eeg (any): The EEG signal sample to be updated.
        '''
        eeg_io = self.eeg_io_router[record]
        if self.in_memory:
            eeg_io.write_eeg_in_memory(eeg=eeg, key=key)
        eeg_io.write_eeg(eeg=eeg, key=key)

    def read_info(self, index: int) -> Dict:
        r'''
        Query the corresponding meta information in the MetaInfoIO according to the the given :obj:`index`.

        In meta infomation, clip_id is required. Specifies the corresponding key of EEG in EEGSginalIO, which can be used to index EEG samples based on :obj:`self.read_eeg(key)`.

        Args:
            index (int): The index of the meta information to be queried.
            
        Returns:
            dict: The meta information.
        '''
        return self.info.iloc[index].to_dict()

    def exist(self, io_path: str) -> bool:
        '''
        Check if the database IO exists.

        Args:
            io_path (str): The path of the database IO.
        Returns:
            bool: True if the database IO exists, otherwise False.
        '''

        return os.path.exists(io_path)

    def __getitem__(self, index: int) -> any:
        info = self.read_info(index)

        eeg_index = str(info['clip_id'])
        eeg_record = str(info['_record_id'])
        eeg = self.read_eeg(eeg_record, eeg_index)

        return eeg, info

    def get_labels(self) -> list:
        '''
        Get the labels of the dataset.

        Returns:
            list: The list of labels.
        '''
        labels = []
        for i in range(len(self)):
            _, label = self.__getitem__(i)
            labels.append(label)
        return labels

    def __len__(self):
        return len(self.info)

    def __copy__(self) -> 'BaseDataset':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)

        result.eeg_io_router, result.info = self.get_pointer(
            io_path=result.io_path,
            io_size=result.io_size,
            io_mode=result.io_mode)
        return result

    @property
    def repr_body(self) -> Dict:
        return {
            'io_path': self.io_path,
            'io_size': self.io_size,
            'io_mode': self.io_mode,
            'in_memory': self.in_memory
        }

    @property
    def repr_tail(self) -> Dict:
        return {'length': self.__len__()}

    def __repr__(self) -> str:
        # init info
        format_string = self.__class__.__name__ + '('
        for i, (k, v) in enumerate(self.repr_body.items()):
            # line end
            if i:
                format_string += ','
            format_string += '\n'
            # str param
            if isinstance(v, str):
                format_string += f"    {k}='{v}'"
            else:
                format_string += f"    {k}={v}"
        format_string += '\n)'
        # other info
        format_string += '\n'
        for i, (k, v) in enumerate(self.repr_tail.items()):
            if i:
                format_string += ', '
            format_string += f'{k}={v}'
        return format_string



import pickle
import os
from typing import Union

import torch
import lmdb


class EEGSignalIO:
    r'''
    A general-purpose, lightweight and efficient EEG signal IO APIs for converting various real-world EEG signal datasets into samples and storing them in the database. Here, we draw on the implementation ideas of industrial-grade application Caffe, and encapsulate a set of EEG signal reading and writing methods based on Lightning Memory-Mapped Database (LMDB), which not only unifies the differences of data types in different databases, but also accelerates the reading of data during training and testing.

    .. code-block:: python

        eeg_io = EEGSignalIO('YOUR_PATH')
        key = eeg_io.write_eeg(np.random.randn(32, 128))
        eeg = eeg_io.read_eeg(key)
        eeg.shape
        >>> (32, 128)
    
    Args:
        io_path (str): Where the database is stored.
        io_size (int, optional): The maximum capacity of the database. It will increase according to the size of the dataset. (default: :obj:`10485760`)
        io_mode (str): Storage mode of EEG signal. When io_mode is set to :obj:`lmdb`, TorchEEG provides an efficient database (LMDB) for storing EEG signals. LMDB may not perform well on limited operating systems, where a file system based EEG signal storage is also provided. When io_mode is set to :obj:`pickle`, pickle-based persistence files are used. (default: :obj:`lmdb`)
    '''
    def __init__(self,
                 io_path: str,
                 io_size: int = 10485760,
                 io_mode: str = 'lmdb') -> None:
        self.io_path = io_path
        self.io_size = io_size
        self.io_mode = io_mode

        assert io_mode in [
            'lmdb', 'pickle'
        ], f'Unsupported io_mode {io_mode}, please choose from [lmdb, pickle]!'

        self._in_memory = None

        if not os.path.exists(self.io_path):
            os.makedirs(self.io_path, exist_ok=True)

    @property
    def write_pointer(self):
        return len(self)

    def __len__(self):
        if self.io_mode == 'pickle':
            return len(os.listdir(self.io_path))

        if self.io_mode == 'lmdb':
            with lmdb.open(path=self.io_path,
                           map_size=self.io_size,
                           lock=False) as env:
                with env.begin() as transaction:
                    return transaction.stat()['entries']

    def write_eeg(self,
                  eeg: Union[any, torch.Tensor],
                  key: Union[str, None] = None) -> str:
        r'''
        Write EEG signal to database.

        Args:
            eeg (any): EEG signal samples to be written into the database.
            key (str, optional): The key of the EEG signal to be inserted, if not specified, it will be an auto-incrementing integer.

        Returns:
            int: The index of written EEG signals in the database.
        '''

        if key is None:
            key = str(self.write_pointer)

        if eeg is None:
            raise RuntimeError(f'Save None to the LMDB with the key {key}!')

        if self.io_mode == 'pickle':
            with open(os.path.join(self.io_path, key), 'wb') as f:
                pickle.dump(eeg, f)
            return key

        if self.io_mode == 'lmdb':
            try_again = False
            with lmdb.open(path=self.io_path,
                           map_size=self.io_size,
                           lock=False,
                           writemap=True,
                           map_async=True) as env:
                try:
                    with env.begin(write=True) as transaction:
                        transaction.put(key.encode(), pickle.dumps(eeg))
                except lmdb.MapFullError:
                    # print(
                    #     f'The current io_size is not enough, and double the LMDB map size to {self.io_size * 2} automatically.'
                    # )
                    self.io_size = self.io_size * 2
                    try_again = True
            if try_again:
                return self.write_eeg(key=key, eeg=eeg)
            return key

    def read_eeg(self, key: str) -> any:
        r'''
        Query the corresponding EEG signal in the database according to the index.

        Args:
            key (str): The index of the EEG signal to be queried.
            
        Returns:
            any: The EEG signal sample.
        '''
        if self.io_mode == 'pickle':
            with open(os.path.join(self.io_path, key), 'rb') as f:
                eeg = pickle.load(f)
            return eeg

        if self.io_mode == 'lmdb':
            with lmdb.open(path=self.io_path,
                           map_size=self.io_size,
                           lock=False) as env:
                with env.begin() as transaction:
                    eeg = transaction.get(key.encode())

                if eeg is None:
                    raise RuntimeError(
                        f'Unable to index the EEG signal sample with key {key}!'
                    )

                return pickle.loads(eeg)

    def keys(self):
        r'''
        Get all keys in the EEGSignalIO.

        Returns:
            list: The list of keys in the EEGSignalIO.
        '''
        if self.io_mode == 'pickle':
            return os.listdir(self.io_path)

        if self.io_mode == 'lmdb':
            with lmdb.open(path=self.io_path,
                           map_size=self.io_size,
                           lock=False) as env:
                with env.begin() as transaction:
                    return [
                        key.decode()
                        for key in transaction.cursor().iternext(keys=True,
                                                                 values=False)
                    ]

    def eegs(self):
        r'''
        Get all EEG signals in the EEGSignalIO.

        Returns:
            list: The list of EEG signals in the EEGSignalIO.
        '''
        return [self.read_eeg(key) for key in self.keys()]

    def to_dict(self):
        r'''
        Convert EEGSignalIO to an in-memory dictionary, where the index of each sample in the database corresponds to the key, and the EEG signal stored in the database corresponds to the value.

        Returns:
            dict: The dict of samples in the EEGSignalIO.
        '''
        return {key: self.read_eeg(key) for key in self.keys()}

    def read_eeg_in_memory(self, key: str) -> any:
        r'''
        Read all the EEGSignalIO into memory, and index the specified EEG signal in memory with the given :obj:`key`.

        .. warning::
            This method will read all the data in EEGSignalIO into memory, which may cause memory overflow. Thus, it is only recommended for fast reading of small-scale datasets.

        Args:
            key (str): The index of the EEG signal to be queried.
            
        Returns:
            any: The EEG signal sample.
        '''
        if self._in_memory is None:
            self._in_memory = self.to_dict()

        return self._in_memory[key]

    def write_eeg_in_memory(self, eeg: any, key: str) -> None:
        r'''
        Write EEG signal to memory.

        .. warning::
            This method will write all the data in EEGSignalIO into memory, which may cause memory overflow. Thus, it is only recommended for fast reading of small-scale datasets.

        Args:
            eeg (any): EEG signal samples to be written into the database.
            key (str): The key of the EEG signal to be inserted, if not specified, it will be an auto-incrementing
        '''
        if self._in_memory is None:
            self._in_memory = self.to_dict()
        
        self._in_memory[key] = eeg


from typing import Tuple

import torch
import torch.nn as nn


class BEncoder(nn.Module):
    r'''
    The variational autoencoder consists of two parts, an encoder, and a decoder. The encoder compresses the input into the latent space. The decoder receives as input the information sampled from the latent space and produces it as similar as possible to ground truth. The latent vector should approach the gaussian distribution supervised by KL divergence based on the variation trick. This class implement the encoder part.

    .. code-block:: python

        encoder = BEncoder(in_channels=4)
        mock_eeg = torch.randn(1, 4, 9, 9)
        mu, logvar = encoder(mock_eeg)

    Args:
        in_channels (int): The feature dimension of each electrode. (default: :obj:`4`)
        grid_size (tuple): Spatial dimensions of grid-like EEG representation. (default: :obj:`(9, 9)`)
        hid_channels (int): The number of hidden nodes in the first convolutional layer, which is also used as the dimension of output mu and var. (default: :obj:`32`)
    '''
    def __init__(self,
                 in_channels: int = 4,
                 grid_size: Tuple[int, int] = (9, 9),
                 hid_channels: int = 64):
        super(BEncoder, self).__init__()

        self.in_channels = in_channels
        self.grid_size = grid_size
        self.hid_channels = hid_channels

        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channels,
                      hid_channels,
                      kernel_size=3,
                      stride=1,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels), nn.LeakyReLU())
        self.conv2 = nn.Sequential(
            nn.Conv2d(hid_channels,
                      hid_channels * 2,
                      kernel_size=3,
                      stride=2,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels * 2),
            nn.LeakyReLU())
        self.conv3 = nn.Sequential(
            nn.Conv2d(hid_channels * 2,
                      hid_channels * 2,
                      kernel_size=3,
                      stride=1,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels * 2),
            nn.LeakyReLU())
        self.conv4 = nn.Sequential(
            nn.Conv2d(hid_channels * 2,
                      hid_channels * 4,
                      kernel_size=3,
                      stride=2,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels * 4),
            nn.LeakyReLU())

        feature_dim = self.feature_dim
        self.fc_mu = nn.Linear(feature_dim, self.hid_channels)
        self.fc_var = nn.Linear(feature_dim, self.hid_channels)

    @property
    def feature_dim(self):
        with torch.no_grad():
            mock_eeg = torch.zeros(1, self.in_channels, *self.grid_size)

            mock_eeg = self.conv1(mock_eeg)
            mock_eeg = self.conv2(mock_eeg)
            mock_eeg = self.conv3(mock_eeg)
            mock_eeg = self.conv4(mock_eeg)

        return mock_eeg.flatten(start_dim=1).shape[-1]

    def forward(self, x: torch.Tensor):
        r'''
        Args:
            x (torch.Tensor): EEG signal representation, the ideal input shape is :obj:`[n, 4, 9, 9]`. Here, :obj:`n` corresponds to the batch size, :obj:`4` corresponds to :obj:`in_channels`, and :obj:`(9, 9)` corresponds to :obj:`grid_size`.

        Returns:
            tuple[2,]: The mean and standard deviation vectors obtained by encoder. The shapes of the feature vectors are all :obj:`[n, 64]`. Here, :obj:`n` corresponds to the batch size, and :obj:`64` corresponds to :obj:`hid_channels`.
        '''
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        x = x.flatten(start_dim=1)
        mu = self.fc_mu(x)
        var = self.fc_var(x)
        return mu, var


class BDecoder(nn.Module):
    r'''
    The variational autoencoder consists of two parts, an encoder, and a decoder. The encoder compresses the input into the latent space. The decoder receives as input the information sampled from the latent space and produces it as similar as possible to ground truth. The latent vector should approach the gaussian distribution supervised by KL divergence based on the variation trick. This class implement the decoder part.

    .. code-block:: python

        encoder = BEncoder(in_channels=4)
        decoder = BDecoder(in_channels=64, out_channels=4)
        mock_eeg = torch.randn(1, 4, 9, 9)
        mu, logvar = encoder(mock_eeg)
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        z = eps * std + mu
        fake_X = decoder(z)

    Args:
        in_channels (int): The input feature dimension (of noise vectors). (default: :obj:`64`)
        out_channels (int): The generated feature dimension of each electrode. (default: :obj:`4`)
        grid_size (tuple): Spatial dimensions of grid-like EEG representation. (default: :obj:`(9, 9)`)
    '''
    def __init__(self,
                 in_channels: int = 64,
                 out_channels: int = 4,
                 grid_size: Tuple[int, int] = (9, 9)):
        super(BDecoder, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.grid_size = grid_size

        self.deproj = nn.Sequential(
            nn.Linear(in_channels, in_channels * 4 * 3 * 3), nn.LeakyReLU())

        self.deconv1 = nn.Sequential(
            nn.ConvTranspose2d(in_channels * 4,
                               in_channels * 2,
                               kernel_size=3,
                               stride=2,
                               padding=1,
                               bias=True), nn.BatchNorm2d(in_channels * 2),
            nn.LeakyReLU())
        self.deconv2 = nn.Sequential(
            nn.ConvTranspose2d(in_channels * 2,
                               in_channels * 2,
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               bias=True), nn.BatchNorm2d(in_channels * 2),
            nn.LeakyReLU())
        self.deconv3 = nn.Sequential(
            nn.ConvTranspose2d(in_channels * 2,
                               in_channels,
                               kernel_size=3,
                               stride=2,
                               padding=1,
                               bias=True), nn.BatchNorm2d(in_channels),
            nn.LeakyReLU())
        self.deconv4 = nn.ConvTranspose2d(in_channels,
                                          out_channels,
                                          kernel_size=3,
                                          stride=1,
                                          padding=1,
                                          bias=True)

    def forward(self, x: torch.Tensor):
        r'''
        Args:
            x (torch.Tensor): Given the mean and standard deviation vectors, the feature vector :obj:`z` obtained using the reparameterization technique. The shapes of the feature vector should be :obj:`[n, 64]`. Here, :obj:`n` corresponds to the batch size, and :obj:`64` corresponds to :obj:`in_channels`.

        Returns:
            torch.Tensor[n, 4, 9, 9]: the decoded results, which should have the same shape as the input noise, i.e., :obj:`[n, 4, 9, 9]`. Here, :obj:`n` corresponds to the batch size, :obj:`4` corresponds to :obj:`in_channels`, and :obj:`(9, 9)` corresponds to :obj:`grid_size`.
        '''
        x = self.deproj(x)
        x = x.view(-1, self.in_channels * 4, 3, 3)
        x = self.deconv1(x)
        x = self.deconv2(x)
        x = self.deconv3(x)
        x = self.deconv4(x)
        return x


import unittest
import os
import random
import shutil

from torcheeg.model_selection import Subcategory
from torcheeg.datasets import M3CVDataset
from torcheeg import transforms


class TestSubcategory(unittest.TestCase):
    def setUp(self):
        shutil.rmtree('./tmp_out/')
        os.mkdir('./tmp_out/')

    def test_k_fold(self):
        io_path = f'./tmp_out/m3cv_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        root_path = './tmp_in/aistudio'

        dataset = M3CVDataset(io_path=io_path, root_path=root_path)

        split_path = f'./tmp_out/split_{"".join(random.sample("zyxwvutsrqponmlkjihgfedcba", 20))}'
        cv = Subcategory(split_path=split_path)

        split_num = 0
        for subdataset in cv.split(dataset):
            # print(len(subdataset))
            split_num += 1
        self.assertEqual(split_num, 13)


if __name__ == '__main__':
    unittest.main()


from typing import Dict, Tuple, Union

from scipy.signal import butter, hann, lfilter

import numpy as np

from ..base_transform import EEGTransform


def butter_bandpass(low_cut, high_cut, fs, order=5):
    nyq = 0.5 * fs
    low = low_cut / nyq
    high = high_cut / nyq
    b, a = butter(order, [low, high], btype='band')
    return b, a


class BandTransform(EEGTransform):
    def __init__(self,
                 sampling_rate: int = 128,
                 order: int = 5,
                 band_dict: Dict[str, Tuple[int, int]] = {
                     "theta": [4, 8],
                     "alpha": [8, 14],
                     "beta": [14, 31],
                     "gamma": [31, 49]
                 },
                 apply_to_baseline: bool = False):
        super(BandTransform, self).__init__(apply_to_baseline=apply_to_baseline)
        self.sampling_rate = sampling_rate
        self.order = order
        self.band_dict = band_dict

    def apply(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        band_list = []
        for low, high in self.band_dict.values():
            c_list = []
            for c in eeg:
                b, a = butter_bandpass(low,
                                       high,
                                       fs=self.sampling_rate,
                                       order=self.order)
                c_list.append(self.opt(lfilter(b, a, c)))
            c_list = np.array(c_list)
            band_list.append(c_list)
        return np.stack(band_list, axis=-1)

    def opt(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        raise NotImplementedError

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'sampling_rate': self.sampling_rate,
                'order': self.order,
                'band_dict': {...}
            })


class BandSignal(BandTransform):
    r'''
    A transform method to split the EEG signal into signals in different sub-bands.

    .. code-block:: python

        transform = BandSignal()
        transform(eeg=np.random.randn(32, 128))['eeg'].shape
        >>> (4, 32, 128)

    Args:
        sampling_rate (int): The original sampling rate of EEG signals in Hz. (default: :obj:`128`)
        order (int): The order of the filter. (default: :obj:`5`)
        band_dict: (dict): Band name and the critical sampling rate or frequencies. By default, the differential entropy of the four sub-bands, theta, alpha, beta and gamma, is calculated. (default: :obj:`{...}`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)
    
    .. automethod:: __call__
    '''
    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.
        Returns:
            np.ndarray[number of electrodes, number of sub-bands]: The differential entropy of several sub-bands for all electrodes.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def opt(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        return eeg

    def apply(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        band_list = []
        for low, high in self.band_dict.values():
            c_list = []
            for c in eeg:
                b, a = butter(self.order, [low, high],
                              fs=self.sampling_rate,
                              btype="band")
                c_list.append(self.opt(lfilter(b, a, c)))
            c_list = np.array(c_list)
            band_list.append(c_list)
        return np.stack(band_list, axis=0)


class BandDifferentialEntropy(BandTransform):
    r'''
    A transform method for calculating the differential entropy of EEG signals in several sub-bands with EEG signals as input. It is a widely accepted differential entropy calculation method by the community, which is often applied to the DEAP and DREAMER datasets. It is relatively easy to understand and has a smaller scale and more gradual changes than the :obj:`BandDifferentialEntropyV1` calculated based on average power spectral density. 

    - Related Paper: Fdez J, Guttenberg N, Witkowski O, et al. Cross-subject EEG-based emotion recognition through neural networks with stratified normalization[J]. Frontiers in neuroscience, 2021, 15: 626277.
    - Related Project: https://github.com/javiferfer/cross-subject-eeg-emotion-recognition-through-nn/

    - Related Paper: Li D, Xie L, Chai B, et al. Spatial-frequency convolutional self-attention network for EEG emotion recognition[J]. Applied Soft Computing, 2022, 122: 108740.
    - Related Project: https://github.com/qeebeast7/SFCSAN/

    In most cases, choosing :obj:`BandDifferentialEntropy` and :obj:`BandDifferentialEntropyV1` does not make much difference. If you have other comments, please feel free to pull request.
    
    .. code-block:: python

        transform = BandDifferentialEntropy()
        transform(eeg=np.random.randn(32, 128))['eeg'].shape
        >>> (32, 4)

    Args:
        sampling_rate (int): The original sampling rate of EEG signals in Hz. (default: :obj:`128`)
        order (int): The order of the filter. (default: :obj:`5`)
        band_dict: (dict): Band name and the critical sampling rate or frequencies. By default, the differential entropy of the four sub-bands, theta, alpha, beta and gamma, is calculated. (default: :obj:`{...}`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)
    
    .. automethod:: __call__
    '''
    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.
        Returns:
            np.ndarray[number of electrodes, number of sub-bands]: The differential entropy of several sub-bands for all electrodes.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def opt(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        return 1 / 2 * np.log2(2 * np.pi * np.e * np.std(eeg))


class BandDifferentialEntropyV1(EEGTransform):
    r'''
    A transform method for calculating the differential entropy of EEG signals in several sub-bands with EEG signals as input. This version calculates the differential entropy based on the relationship between the differential entropy and the average power spectral density, which is identical to the processing of the SEED dataset.

    - Related Paper: Shi L C, Jiao Y Y, Lu B L. Differential entropy feature for EEG-based vigilance estimation[C]//2013 35th Annual International Conference of the IEEE Engineering in Medicine and Biology Society (EMBC). IEEE, 2013: 6627-6630.
    - Related Project: https://github.com/ziyujia/Signal-feature-extraction_DE-and-PSD

    .. code-block:: python

        transform = BandDifferentialEntropyV1()
        transform(eeg=np.random.randn(32, 128))['eeg'].shape
        >>> (32, 4)

    Args:
        sampling_rate (int): The original sampling rate of EEG signals in Hz. (default: :obj:`128`)
        order (int): The order of the filter. (default: :obj:`5`)
        band_dict: (dict): Band name and the critical sampling rate or frequencies. By default, the differential entropy of the four sub-bands, theta, alpha, beta and gamma, is calculated. (default: :obj:`{...}`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)
    
    .. automethod:: __call__
    '''
    def __init__(self,
                 sampling_rate: int = 128,
                 fft_n: int = None,
                 num_window: int = 1,
                 band_dict: Dict[str, Tuple[int, int]] = {
                     "theta": [4, 8],
                     "alpha": [8, 14],
                     "beta": [14, 31],
                     "gamma": [31, 49]
                 },
                 apply_to_baseline: bool = False):
        super(BandDifferentialEntropyV1,
              self).__init__(apply_to_baseline=apply_to_baseline)
        self.sampling_rate = sampling_rate

        if fft_n is None:
            fft_n = self.sampling_rate

        self.fft_n = fft_n
        self.num_window = num_window
        self.band_dict = band_dict

    def apply(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        _, chunk_size = eeg.shape
        point_per_window = int(chunk_size // self.num_window)

        band_list = []

        for window_index in range(self.num_window):
            start_index, end_index = point_per_window * window_index, point_per_window * (
                window_index + 1)
            window_data = eeg[:, start_index:end_index]
            hdata = window_data * hann(point_per_window)
            fft_data = np.fft.fft(hdata, n=self.fft_n)
            energy_graph = np.abs(fft_data[:, 0:int(self.fft_n / 2)])

            for _, band in enumerate(self.band_dict.values()):
                start_index = int(
                    np.floor(band[0] / self.sampling_rate * self.fft_n))
                end_index = int(
                    np.floor(band[1] / self.sampling_rate * self.fft_n))
                band_ave_psd = np.mean(energy_graph[:, start_index -
                                                    1:end_index]**2,
                                       axis=1)
                # please refer to # https://github.com/ziyujia/Signal-feature-extraction_DE-and-PSD/blob/master/DE_PSD.py, which consider the relationship between DE and PSD to calculate DE.
                band_de = np.log2(100 * band_ave_psd)
                band_list.append(band_de)

        return np.stack(band_list, axis=-1)

    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.
        Returns:
            np.ndarray[number of electrodes, number of sub-bands]: The power spectral density of several sub-bands for all electrodes.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'sampling_rate': self.sampling_rate,
                'fft_n': self.fft_n,
                'num_window': self.num_window,
                'band_dict': {...}
            })


class BandPowerSpectralDensity(EEGTransform):
    r'''
    A transform method for calculating the power spectral density of EEG signals in several sub-bands with EEG signals as input.

    .. code-block:: python

        transform = BandPowerSpectralDensity()
        transform(eeg=np.random.randn(32, 128))['eeg'].shape
        >>> (32, 4)

    Args:
        sampling_rate (int): The sampling rate of EEG signals in Hz. (default: :obj:`128`)
        fft_n (int): Computes the one-dimensional n-point discrete Fourier Transform (DFT) with the efficient Fast Fourier Transform (FFT) algorithm. If set to None, it will automatically match sampling_rate. (default: :obj:`None`)
        num_window (int): Welch's method computes an estimate of the power spectral density by dividing the data into non-overlapping segments, where the num_window denotes the number of windows. (default: :obj:`1`)
        order (int): The order of the filter. (default: :obj:`5`)
        band_dict: (dict): Band name and the critical sampling rate or frequencies. By default, the power spectral density of the four sub-bands, theta, alpha, beta and gamma, is calculated. (default: :obj:`{...}`)
        apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)

    .. automethod:: __call__
    '''
    def __init__(self,
                 sampling_rate: int = 128,
                 fft_n: int = None,
                 num_window: int = 1,
                 band_dict: Dict[str, Tuple[int, int]] = {
                     "theta": [4, 8],
                     "alpha": [8, 14],
                     "beta": [14, 31],
                     "gamma": [31, 49]
                 },
                 apply_to_baseline: bool = False):
        super(BandPowerSpectralDensity,
              self).__init__(apply_to_baseline=apply_to_baseline)
        self.sampling_rate = sampling_rate

        if fft_n is None:
            fft_n = self.sampling_rate

        self.fft_n = fft_n
        self.num_window = num_window
        self.band_dict = band_dict

    def apply(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        _, chunk_size = eeg.shape
        point_per_window = int(chunk_size // self.num_window)

        band_list = []

        for window_index in range(self.num_window):
            start_index, end_index = point_per_window * window_index, point_per_window * (
                window_index + 1)
            window_data = eeg[:, start_index:end_index]
            hdata = window_data * hann(point_per_window)
            fft_data = np.fft.fft(hdata, n=self.fft_n)
            energy_graph = np.abs(fft_data[:, 0:int(self.fft_n / 2)])

            for _, band in enumerate(self.band_dict.values()):
                start_index = int(
                    np.floor(band[0] / self.sampling_rate * self.fft_n))
                end_index = int(
                    np.floor(band[1] / self.sampling_rate * self.fft_n))
                band_ave_psd = np.mean(energy_graph[:, start_index -
                                                    1:end_index]**2,
                                       axis=1)

                band_list.append(band_ave_psd)

        return np.stack(band_list, axis=-1)

    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.
        Returns:
            np.ndarray[number of electrodes, number of sub-bands]: The power spectral density of several sub-bands for all electrodes.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'sampling_rate': self.sampling_rate,
                'fft_n': self.fft_n,
                'num_window': self.num_window,
                'band_dict': {...}
            })


class BandMeanAbsoluteDeviation(BandTransform):
    r'''
    A transform method for calculating the mean absolute deviation of EEG signals in several sub-bands with EEG signals as input.

    .. code-block:: python

        transform = BandMeanAbsoluteDeviation()
        transform(eeg=np.random.randn(32, 128))['eeg'].shape
        >>> (32, 4)

    Args:
        sampling_rate (int): The original sampling rate of EEG signals in Hz. (default: :obj:`128`)
        order (int): The order of the filter. (default: :obj:`5`)
        band_dict: (dict): Band name and the critical sampling rate or frequencies. By default, the mean absolute deviation of the four sub-bands, theta, alpha, beta and gamma, is calculated. (default: :obj:`{...}`)
    
    .. automethod:: __call__
    '''
    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.
        Returns:
            np.ndarray[number of electrodes, number of sub-bands]: The mean absolute deviation of several sub-bands for all electrodes.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def opt(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        return np.mean(np.abs(eeg - np.mean(eeg)))


class BandKurtosis(BandTransform):
    r'''
    A transform method for calculating the kurtosis of EEG signals in several sub-bands with EEG signals as input.

    .. code-block:: python

        transform = BandKurtosis()
        transform(eeg=np.random.randn(32, 128))['eeg'].shape
        >>> (32, 4)

    Args:
        sampling_rate (int): The original sampling rate of EEG signals in Hz. (default: :obj:`128`)
        order (int): The order of the filter. (default: :obj:`5`)
        band_dict: (dict): Band name and the critical sampling rate or frequencies. By default, the kurtosis of the four sub-bands, theta, alpha, beta and gamma, is calculated. (default: :obj:`{...}`)
    
    .. automethod:: __call__
    '''
    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.
        Returns:
            np.ndarray[number of electrodes, number of sub-bands]: The kurtosis of several sub-bands for all electrodes.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def opt(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        n = len(eeg)
        ave1 = 0.0
        ave2 = 0.0
        ave4 = 0.0
        for eeg in eeg:
            ave1 += eeg
            ave2 += eeg**2
            ave4 += eeg**4
        ave1 /= n
        ave2 /= n
        ave4 /= n
        sigma = np.sqrt(ave2 - ave1**2)
        return ave4 / (sigma**4)


class BandSkewness(BandTransform):
    r'''
    A transform method for calculating the skewness of EEG signals in several sub-bands with EEG signals as input.

    .. code-block:: python

        transform = BandSkewness()
        transform(eeg=np.random.randn(32, 128))['eeg'].shape
        >>> (32, 4)

    Args:
        sampling_rate (int): The original sampling rate of EEG signals in Hz. (default: :obj:`128`)
        order (int): The order of the filter. (default: :obj:`5`)
        band_dict: (dict): Band name and the critical sampling rate or frequencies. By default, the skewness of the four sub-bands, theta, alpha, beta and gamma, is calculated. (default: :obj:`{...}`)

    .. automethod:: __call__
    '''
    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.
        Returns:
            np.ndarray[number of electrodes, number of sub-bands]: The skewness of several sub-bands for all electrodes.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def opt(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        n = len(eeg)
        ave1 = 0.0
        ave2 = 0.0
        ave3 = 0.0
        for eeg in eeg:
            ave1 += eeg
            ave2 += eeg**2
            ave3 += eeg**3
        ave1 /= n
        ave2 /= n
        ave3 /= n
        sigma = np.sqrt(ave2 - ave1**2)
        return (ave3 - 3 * ave1 * sigma**2 - ave1**3) / (sigma**3)



from typing import Dict, Union

import io
import pywt
import numpy as np
import matplotlib.pyplot as plt

from ..base_transform import EEGTransform


class CWTSpectrum(EEGTransform):
    r'''
        A transform method to convert EEG signals of each channel into spectrograms using wavelet transform.

        .. code-block:: python

            transform = CWTSpectrum()
            transform(eeg=np.random.randn(32, 1000))['eeg'].shape
            >>> (32, 128, 1000)

        Part of the existing work uses :obj:`Resize` to warp the output spectrum to a specified size suitable for CNN processing.

        .. code-block:: python

            transform = Compose([
                CWTSpectrum(),
                ToTensor(),
                Resize([260, 260])
            ])
            transform(eeg=np.random.randn(32, 1000))['eeg'].shape
            >>> (32, 128, 1000)

        When contourf is set to True, a spectrogram of filled contours will be generated for each channel and converted to np.ndarray and returned. This option is usually used for single-channel analysis or visualization of a single channel.

        .. code-block:: python

            transform = CWTSpectrum(contourf=True)
            transform(eeg=np.random.randn(32, 1000))['eeg'].shape
            >>> (32, 480, 640, 4)

        Args:
            sampling_rate (int): The sampling period for the frequencies output in Hz. (default: :obj:`128`)
            wavelet (str): Wavelet to use. Options include: cgau1, cgau2, cgau3, cgau4, cgau5, cgau6, cgau7, cgau8, cmor, fbsp, gaus1, gaus2 , gaus3, gaus4, gaus5, gaus6, gaus7, gaus8, mexh, morl, shan. (default: :obj:`'morl'`)
            total_scale: (int): The total wavelet scales to use. (default: :obj:`128`)
            contourf: (bool): Whether to output the np.ndarray corresponding to the image with content of filled contours. (default: :obj:`False`)
            apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)
        
        .. automethod:: __call__
    '''
    def __init__(self,
                 sampling_rate: int = 250,
                 wavelet: str = 'morl',
                 total_scale: int = 128,
                 contourf: bool = False,
                 apply_to_baseline: bool = False):
        super(CWTSpectrum, self).__init__(apply_to_baseline=apply_to_baseline)
        self.sampling_rate = sampling_rate
        self.wavelet = wavelet
        self.total_scale = total_scale
        self.contourf = contourf

        fc = pywt.central_frequency(wavelet)
        cparam = 2 * fc * total_scale
        self.scales = cparam / np.arange(1, self.total_scale + 1)

    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.
        Returns:
            np.ndarray[number of electrodes, ...]: The spectrograms based on the wavelet transform for all electrodes. If contourf=False, the output shape is [number of electrodes, total_scale, number of data points]. Otherwise, the output shape is [number of electrodes, height of image, width of image of image, 4], where 4 represents the four channels of the image colors.
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def apply(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        channel_list = []
        for channel in eeg:
            channel_list.append(self.opt(channel))
        channel_list = np.array(channel_list)
        return np.array(channel_list)

    def opt(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        t = np.arange(0,
                      len(eeg) / self.sampling_rate, 1.0 / self.sampling_rate)
        [cwtmatr, frequencies] = pywt.cwt(eeg, self.scales, self.wavelet,
                                          1.0 / self.sampling_rate)

        if self.contourf:
            fig = plt.figure()

            plt.xticks([])
            plt.yticks([])
            plt.axis('off')

            plt.gca().xaxis.set_major_locator(plt.NullLocator())
            plt.gca().yaxis.set_major_locator(plt.NullLocator())
            plt.subplots_adjust(top=1,
                                bottom=0,
                                left=0,
                                right=1,
                                hspace=0,
                                wspace=0)
            plt.margins(0, 0)

            plt.contourf(t, frequencies, abs(cwtmatr))

            with io.BytesIO() as buf:
                fig.savefig(buf, format='raw')
                buf.seek(0)
                img_cwtmatr = np.frombuffer(buf.getvalue(), dtype=np.uint8)
                w, h = fig.canvas.get_width_height()
                img_cwtmatr = img_cwtmatr.reshape((int(h), int(w), -1))

            return img_cwtmatr

        return cwtmatr

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'sampling_rate': self.sampling_rate,
                'wavelet': self.wavelet,
                'total_scale': self.total_scale,
                'contourf': self.contourf
            })


class DWTDecomposition(EEGTransform):
    r'''    
        Splitting the EEG signal from each electrode into two functions using wavelet decomposition.

        .. code-block:: python

            transform = DWTDecomposition()
            transform(eeg=np.random.randn(32, 1000))['eeg'].shape
            >>> (32, 500)

        Args:
            apply_to_baseline: (bool): Whether to act on the baseline signal at the same time, if the baseline is passed in when calling. (default: :obj:`False`)
        
        .. automethod:: __call__
    '''
    def __init__(self, apply_to_baseline: bool = False):
        super(DWTDecomposition,
              self).__init__(apply_to_baseline=apply_to_baseline)

    def __call__(self,
                 *args,
                 eeg: np.ndarray,
                 baseline: Union[np.ndarray, None] = None,
                 **kwargs) -> Dict[str, np.ndarray]:
        r'''
        Args:
            eeg (np.ndarray): The input EEG signals in shape of [number of electrodes, number of data points].
            baseline (np.ndarray, optional) : The corresponding baseline signal, if apply_to_baseline is set to True and baseline is passed, the baseline signal will be transformed with the same way as the experimental signal.
        Returns:
            np.ndarray[number of electrodes, 2, number of data points / 2]: EEG signal after wavelet decomposition, where 2 corresponds to the two functions of the wavelet decomposition, and number of data points / 2 represents the length of each component
        '''
        return super().__call__(*args, eeg=eeg, baseline=baseline, **kwargs)

    def apply(self, eeg: np.ndarray, **kwargs) -> np.ndarray:
        return np.stack(pywt.dwt(eeg, 'haar'), axis=0)


try:
    import torch_geometric
except ImportError as e:
    raise ImportError(
        "To run the torcheeg with graph neural networks and related transforms, you need to install `torch_geometric`. Please refer to `https://pytorch-geometric.readthedocs.io/en/latest/notes/installation.html`."
    ) from e
    
from .visualize import plot_graph


from typing import Tuple

import torch
import torch.nn as nn


class BCEncoder(nn.Module):
    r'''
    TorchEEG provides an EEG feature encoder based on CNN architecture and CVAE for generating EEG grid representations of different frequency bands based on a given class label. In particular, the expected labels are additionally provided to guide the encoder to derive the mean and standard deviation vectors of the given expected labels and input data.

    - Related Project: https://github.com/timbmg/VAE-CVAE-MNIST/blob/master/models.py

    .. code-block:: python

        encoder = BCEncoder(in_channels=4, num_classes=3)
        y = torch.randint(low=0, high=3, size=(1,))
        mock_eeg = torch.randn(1, 4, 9, 9)
        mu, logvar = encoder(mock_eeg, y)

    Args:
        in_channels (int): The feature dimension of each electrode. (default: :obj:`4`)
        grid_size (tuple): Spatial dimensions of grid-like EEG representation. (default: :obj:`(9, 9)`)
        hid_channels (int): The number of hidden nodes in the first convolutional layer, which is also used as the dimension of output mu and logvar. (default: :obj:`32`)
        num_classes (int): The number of classes. (default: :obj:`2`)
    '''
    def __init__(self,
                 in_channels: int = 4,
                 grid_size: Tuple[int, int] = (9, 9),
                 hid_channels: int = 64,
                 num_classes: int = 3):
        super(BCEncoder, self).__init__()

        self.in_channels = in_channels
        self.grid_size = grid_size
        self.hid_channels = hid_channels
        self.num_classes = num_classes

        self.label_embeding = nn.Embedding(
            num_classes, in_channels * grid_size[0] * grid_size[1])
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channels * 2,
                      hid_channels,
                      kernel_size=3,
                      stride=1,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels), nn.LeakyReLU())
        self.conv2 = nn.Sequential(
            nn.Conv2d(hid_channels,
                      hid_channels * 2,
                      kernel_size=3,
                      stride=2,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels * 2),
            nn.LeakyReLU())
        self.conv3 = nn.Sequential(
            nn.Conv2d(hid_channels * 2,
                      hid_channels * 2,
                      kernel_size=3,
                      stride=1,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels * 2),
            nn.LeakyReLU())
        self.conv4 = nn.Sequential(
            nn.Conv2d(hid_channels * 2,
                      hid_channels * 4,
                      kernel_size=3,
                      stride=2,
                      padding=1,
                      bias=True), nn.BatchNorm2d(hid_channels * 4),
            nn.LeakyReLU())

        feature_dim = self.feature_dim()
        self.fc_mu = nn.Linear(feature_dim, self.hid_channels)
        self.fc_var = nn.Linear(feature_dim, self.hid_channels)

    def feature_dim(self):
        with torch.no_grad():
            mock_y = torch.randint(low=0, high=self.num_classes, size=(1, ))
            mock_eeg = torch.zeros(1, self.in_channels, *self.grid_size)

            label_emb = self.label_embeding(mock_y)
            label_emb = label_emb.reshape(mock_eeg.shape)
            mock_eeg = torch.cat([mock_eeg, label_emb], dim=1)

            mock_eeg = self.conv1(mock_eeg)
            mock_eeg = self.conv2(mock_eeg)
            mock_eeg = self.conv3(mock_eeg)
            mock_eeg = self.conv4(mock_eeg)

        return mock_eeg.flatten(start_dim=1).shape[-1]

    def forward(self, x: torch.Tensor, y: torch.Tensor):
        r'''
        Args:
            x (torch.Tensor): EEG signal representation, the ideal input shape is :obj:`[n, 4, 9, 9]`. Here, :obj:`n` corresponds to the batch size, :obj:`4` corresponds to :obj:`in_channels`, and :obj:`(9, 9)` corresponds to :obj:`grid_size`.
            y (torch.Tensor): Category labels (int) for a batch of samples The shape should be :obj:`[n,]`. Here, :obj:`n` corresponds to the batch size.

        Returns:
            tuple[2,]: The mean and standard deviation vectors obtained by encoder. The shapes of the feature vectors are all :obj:`[n, 64]`. Here, :obj:`n` corresponds to the batch size, and :obj:`64` corresponds to :obj:`hid_channels`.
        '''
        label_emb = self.label_embeding(y)
        label_emb = label_emb.reshape(x.shape)
        x = torch.cat([x, label_emb], dim=1)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        x = x.flatten(start_dim=1)
        mu = self.fc_mu(x)
        var = self.fc_var(x)
        return mu, var


class BCDecoder(nn.Module):
    r'''
    TorchEEG provides an EEG feature decoder based on CNN architecture and CVAE for generating EEG grid representations of different frequency bands based on a given class label. In particular, the expected labels are additionally provided to guide the decoder to reconstruct samples of the specified class.

    - Related Project: https://github.com/timbmg/VAE-CVAE-MNIST/blob/master/models.py

    .. code-block:: python

        encoder = BCEncoder(in_channels=4, num_classes=3)
        decoder = BCDecoder(in_channels=64, out_channels=4, num_classes=3)
        y = torch.randint(low=0, high=3, size=(1,))
        mock_eeg = torch.randn(1, 4, 9, 9)
        mu, logvar = encoder(mock_eeg, y)
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        z = eps * std + mu
        fake_X = decoder(z, y)

    Args:
        in_channels (int): The input feature dimension (of noise vectors). (default: :obj:`64`)
        out_channels (int): The generated feature dimension of each electrode. (default: :obj:`4`)
        grid_size (tuple): Spatial dimensions of grid-like EEG representation. (default: :obj:`(9, 9)`)
    '''
    def __init__(self,
                 in_channels: int = 64,
                 out_channels: int = 4,
                 grid_size: Tuple[int, int] = (9, 9),
                 num_classes: int = 3):
        super(BCDecoder, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.grid_size = grid_size
        self.num_classes = num_classes

        self.label_embeding = nn.Embedding(num_classes, in_channels)
        self.deproj = nn.Sequential(
            nn.Linear(in_channels * 2, in_channels * 4 * 3 * 3), nn.LeakyReLU())

        self.deconv1 = nn.Sequential(
            nn.ConvTranspose2d(in_channels * 4,
                               in_channels * 2,
                               kernel_size=3,
                               stride=2,
                               padding=1,
                               bias=True), nn.BatchNorm2d(in_channels * 2),
            nn.LeakyReLU())
        self.deconv2 = nn.Sequential(
            nn.ConvTranspose2d(in_channels * 2,
                               in_channels * 2,
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               bias=True), nn.BatchNorm2d(in_channels * 2),
            nn.LeakyReLU())
        self.deconv3 = nn.Sequential(
            nn.ConvTranspose2d(in_channels * 2,
                               in_channels,
                               kernel_size=3,
                               stride=2,
                               padding=1,
                               bias=True), nn.BatchNorm2d(in_channels),
            nn.LeakyReLU())
        self.deconv4 = nn.ConvTranspose2d(in_channels,
                                          out_channels,
                                          kernel_size=3,
                                          stride=1,
                                          padding=1,
                                          bias=True)

    def forward(self, x: torch.Tensor, y: torch.Tensor):
        r'''
        Args:
            x (torch.Tensor): Given the mean and standard deviation vectors, the feature vector :obj:`z` obtained using the reparameterization technique. The shapes of the feature vector should be :obj:`[n, 64]`. Here, :obj:`n` corresponds to the batch size, and :obj:`64` corresponds to :obj:`in_channels`.
            y (torch.Tensor): Category labels (int) for a batch of samples The shape should be :obj:`[n,]`. Here, :obj:`n` corresponds to the batch size.

        Returns:
            torch.Tensor[n, 4, 9, 9]: the decoded results, which should have the same shape as the input noise, i.e., :obj:`[n, 4, 9, 9]`. Here, :obj:`n` corresponds to the batch size, :obj:`4` corresponds to :obj:`in_channels`, and :obj:`(9, 9)` corresponds to :obj:`grid_size`.
        '''
        label_emb = self.label_embeding(y)
        x = torch.cat([x, label_emb], dim=-1)
        x = self.deproj(x)
        x = x.view(-1, self.in_channels * 4, 3, 3)
        x = self.deconv1(x)
        x = self.deconv2(x)
        x = self.deconv3(x)
        x = self.deconv4(x)
        return x